searchData={"items":[{"type":"module","title":"Buckets","doc":"Cloud-agnostic file storage for Elixir with Phoenix integration.\n\nBuckets provides a unified API for working with file storage across different\ncloud providers. Whether you're using local filesystem storage for development,\nGoogle Cloud Storage, Amazon S3, or other S3-compatible providers, Buckets\noffers a consistent interface with powerful features.","ref":"Buckets.html"},{"type":"module","title":"Features - Buckets","doc":"- **Multiple Storage Adapters** - Support for filesystem, Google Cloud Storage,\n  Amazon S3, and S3-compatible providers (Cloudflare R2, DigitalOcean Spaces, Tigris)\n- **Direct Uploads** - Upload files directly from browsers to cloud storage,\n  bypassing your Phoenix server\n- **Signed URLs** - Generate time-limited, secure URLs for private files\n- **LiveView Integration** - Seamless integration with Phoenix LiveView's\n  file upload functionality\n- **Dynamic Configuration** - Switch storage providers at runtime for\n  multi-tenant applications\n- **Development Tools** - Built-in router for local file uploads/downloads\n  during development\n- **Telemetry** - Comprehensive instrumentation for monitoring and debugging","ref":"Buckets.html#module-features"},{"type":"module","title":"Quick Start - Buckets","doc":"1. Add `buckets` to your dependencies:\n\n    ```elixir\n    def deps do\n      [{:buckets, \"~> 1.0.0-rc.2\"}]\n    end\n    ```\n\n2. Create a Cloud module:\n\n    ```elixir\n    defmodule MyApp.Cloud do\n      use Buckets.Cloud, otp_app: :my_app\n    end\n    ```\n\n3. Configure your adapter:\n\n    ```elixir\n    # config/dev.exs\n    config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.Volume,\n      bucket: \"priv/uploads\",\n      base_url: \"http://localhost:4000\"\n    ```\n\n4. Upload files:\n\n    ```elixir\n    # From a Plug.Upload or Phoenix.LiveView.UploadEntry\n    {:ok, object} = MyApp.Cloud.insert(upload)\n\n    # From a file path\n    object = Buckets.Object.from_file(\"photo.jpg\")\n    {:ok, stored} = MyApp.Cloud.insert(object)\n\n    # Generate a signed URL\n    {:ok, url} = MyApp.Cloud.url(stored, expires: 3600)\n    ```","ref":"Buckets.html#module-quick-start"},{"type":"module","title":"Storage Adapters - Buckets","doc":"Buckets includes these built-in adapters:\n\n- `Buckets.Adapters.Volume` - Local filesystem storage\n- `Buckets.Adapters.S3` - Amazon S3 and S3-compatible services\n- `Buckets.Adapters.GCS` - Google Cloud Storage\n\nSee the adapter modules for specific configuration options.","ref":"Buckets.html#module-storage-adapters"},{"type":"module","title":"Next Steps - Buckets","doc":"- See `Buckets.Cloud` for the high-level API\n- See `Buckets.Object` for working with file objects\n- Read the [Getting Started](guides/introduction/getting-started.html) guide\n- Learn about [Direct Uploads with LiveView](guides/howtos/direct-uploads-liveview.html)","ref":"Buckets.html#module-next-steps"},{"type":"function","title":"Buckets.delete/2","doc":"Delegates a `delete/2` function call to the configured `:adapter`.","ref":"Buckets.html#delete/2"},{"type":"function","title":"Buckets.get/2","doc":"Delegates a `get/2` function call to the configured `:adapter`.","ref":"Buckets.html#get/2"},{"type":"function","title":"Buckets.put/3","doc":"Delegates a `put/3` function call to the configured `:adapter`.","ref":"Buckets.html#put/3"},{"type":"function","title":"Buckets.url/2","doc":"Delegates a `url/2` function call to the configured `:adapter`.","ref":"Buckets.html#url/2"},{"type":"behaviour","title":"Buckets.Cloud","doc":"Defines a cloud.\n\nA cloud manages the movement of files and data between your application and\na remote bucket for persistent storage.\n\nWhen used, it expects `:otp_app` as an option:\n\n    defmodule MyApp.Cloud do\n      use Buckets.Cloud, otp_app: :my_app\n    end\n\nConfiguration is fetched from the application config, using a combination of\n`:otp_app` and the module that you defined:\n\n    config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.Volume,\n      bucket: \"tmp/buckets_volume\",\n      base_url: \"http://localhost:4000\"\n\nEach Cloud module corresponds to a single storage backend, similar to how\nEcto.Repo modules correspond to a single database. For multi-cloud applications,\ndefine multiple Cloud modules:\n\n    defmodule MyApp.VolumeCloud do\n      use Buckets.Cloud, otp_app: :my_app\n    end\n\n    defmodule MyApp.GCSCloud do\n      use Buckets.Cloud, otp_app: :my_app\n    end\n\n    defmodule MyApp.S3Cloud do\n      use Buckets.Cloud, otp_app: :my_app\n    end\n\nYou may also specify config dynamically at runtime, using the `:config` opt\nwhere it is supported.","ref":"Buckets.Cloud.html"},{"type":"behaviour","title":"Supervision - Buckets.Cloud","doc":"Cloud modules start a supervisor that manages any background processes required\nby the configured adapter. Some adapters (like GCS) need authentication servers,\nwhile others (like Volume, S3) don't need any supervised processes.\n\n**Only add Cloud modules to your supervision tree if they need background processes.**\nIf you add a Cloud module that doesn't need supervision (Volume, S3), you'll see\na warning message suggesting you remove it to avoid unnecessary overhead.\n\n    # Only needed for adapters that require background processes (like GCS)\n    children = [\n      MyApp.GCSCloud  # GCS needs auth servers\n      # MyApp.VolumeCloud - Not needed, would show warning\n      # MyApp.S3Cloud - Not needed, would show warning\n    ]","ref":"Buckets.Cloud.html#module-supervision"},{"type":"behaviour","title":"Dynamic Configuration - Buckets.Cloud","doc":"For multi-tenant applications where cloud configurations are determined at runtime,\nevery Cloud module supports dynamic configuration using the process dictionary,\nsimilar to Ecto's dynamic repositories.","ref":"Buckets.Cloud.html#module-dynamic-configuration"},{"type":"behaviour","title":"Usage - Buckets.Cloud","doc":"There are two ways to use dynamic configuration:\n\n#### 1. Scoped Configuration (like Ecto transactions)\n\nUse `with_config/2` for temporary configuration:\n\n    # Define the runtime configuration\n    config = [\n      adapter: Buckets.Adapters.S3,\n      bucket: \"user-bucket\",\n      access_key_id: \"AKIA...\",\n      secret_access_key: \"secret...\",\n      region: \"us-east-1\"\n    ]\n\n    # Execute operations with the dynamic config\n    {:ok, object} = MyApp.Cloud.with_config(config, fn ->\n      MyApp.Cloud.insert(\"file.pdf\")\n      MyApp.Cloud.insert(\"another.pdf\")  # Same config\n    end)\n\n#### 2. Process-Local Configuration (like Ecto.Repo.put_dynamic_repo)\n\nUse `put_dynamic_config/1` for persistent configuration in the current process:\n\n    # Set dynamic config for this process\n    :ok = MyApp.Cloud.put_dynamic_config([\n      adapter: Buckets.Adapters.GCS,\n      bucket: \"tenant-specific-bucket\",\n      service_account_credentials: tenant.credentials\n    ])\n\n    # All subsequent operations use the dynamic config\n    {:ok, object1} = MyApp.Cloud.insert(\"file1.pdf\")\n    {:ok, object2} = MyApp.Cloud.insert(\"file2.pdf\")","ref":"Buckets.Cloud.html#module-usage"},{"type":"behaviour","title":"Auth Server Management - Buckets.Cloud","doc":"Auth servers (for GCS) are automatically started and cached per-process as needed.\nYou don't need any special configuration or supervision setup for dynamic clouds.","ref":"Buckets.Cloud.html#module-auth-server-management"},{"type":"callback","title":"Buckets.Cloud.delete/1","doc":"Deletes a `Buckets.Object` permanently.","ref":"Buckets.Cloud.html#c:delete/1"},{"type":"callback","title":"Buckets.Cloud.delete!/1","doc":"Same as `c:delete/1` but returns the object or raises if there is an error.","ref":"Buckets.Cloud.html#c:delete!/1"},{"type":"callback","title":"Buckets.Cloud.insert/2","doc":"Inserts a `Buckets.Object` or a file from a path into a bucket.","ref":"Buckets.Cloud.html#c:insert/2"},{"type":"callback","title":"Buckets.Cloud.insert!/2","doc":"Same as `c:insert/2` but returns the object or raises if there is an error.","ref":"Buckets.Cloud.html#c:insert!/2"},{"type":"callback","title":"Buckets.Cloud.live_upload/2","doc":"Returns a map to be used as configuration for a LiveView live upload. The configuration\ncontains a signed URL that permits the upload to a remote bucket. Requires that an `:uploader`\nis configured for the location that the file is being uploaded to.","ref":"Buckets.Cloud.html#c:live_upload/2"},{"type":"callback","title":"Buckets.Cloud.live_upload!/2","doc":"Same as `c:live_upload/2` but returns the upload config or raises if there is an error.","ref":"Buckets.Cloud.html#c:live_upload!/2"},{"type":"callback","title":"Buckets.Cloud.load/2","doc":"Loads the data for a `Buckets.Object`, placing it in memory by default. It will load data\nlazily, doing nothing if data is already present.","ref":"Buckets.Cloud.html#c:load/2"},{"type":"callback","title":"Options - Buckets.Cloud.load/2","doc":"* `:to` - A location to store the loaded data to on disk, if this is preferred over\n      loading to memory. May be a path, `:tmp` (to load to the configured tmp dir), or\n      `{:tmp, path}` (to load to a path in the configured tmp dir).\n    * `:force` - If data is already present, first unload it with `c:unload/1` before loading\n      new data. Warning: if data is stored in a file, it will be deleted.","ref":"Buckets.Cloud.html#c:load/2-options"},{"type":"callback","title":"Buckets.Cloud.load!/2","doc":"Same as `c:load/2` but returns the object or raises if there is an error.","ref":"Buckets.Cloud.html#c:load!/2"},{"type":"callback","title":"Buckets.Cloud.normalize_filename/1","doc":"An overridable function that processes filenames before encoding them in a remote path.\n\nBy default, replaces whitespace with `\"_\"` and removes all non-alphanumeric characters.","ref":"Buckets.Cloud.html#c:normalize_filename/1"},{"type":"callback","title":"Buckets.Cloud.read/1","doc":"Reads the data for `Buckets.Object`, preferring local data first and fetching from\nthe remote bucket if needed.","ref":"Buckets.Cloud.html#c:read/1"},{"type":"callback","title":"Buckets.Cloud.read!/1","doc":"Same as `c:read/1` but returns the data or raises if there is an error.","ref":"Buckets.Cloud.html#c:read!/1"},{"type":"callback","title":"Buckets.Cloud.tmp_dir/0","doc":"An overridable function that specifies the temporary directory in which objects are stored.\n\nDefaults to `System.tmp_dir!()`.","ref":"Buckets.Cloud.html#c:tmp_dir/0"},{"type":"callback","title":"Buckets.Cloud.unload/1","doc":"Unloads the data for a `Buckets.Object`. If the data is stored in a local file, the file will\nbe deleted.","ref":"Buckets.Cloud.html#c:unload/1"},{"type":"callback","title":"Buckets.Cloud.url/1","doc":"Returns a SignedURL struct for a `Buckets.Object`.","ref":"Buckets.Cloud.html#c:url/1"},{"type":"callback","title":"Buckets.Cloud.url!/2","doc":"Same as `c:url/1` but returns the SignedURL raises if there is an error.","ref":"Buckets.Cloud.html#c:url!/2"},{"type":"module","title":"Buckets.Location","doc":"Represents a storage location for a Buckets.Object.\n\nThe config field can be either:\n- A keyword list with adapter configuration\n- A Cloud module atom that implements the Buckets.Cloud behaviour","ref":"Buckets.Location.html"},{"type":"function","title":"Buckets.Location.get_config/1","doc":"Gets the configuration from a Location.\n\nIf the config is a keyword list, returns it as-is.\nIf the config is a Cloud module, calls the module's config/0 function.","ref":"Buckets.Location.html#get_config/1"},{"type":"function","title":"Buckets.Location.new/2","doc":"Creates a new Location struct with the given path and configuration.","ref":"Buckets.Location.html#new/2"},{"type":"function","title":"Parameters - Buckets.Location.new/2","doc":"- `path` - The storage path for the object (e.g., \"uploads/123/file.pdf\")\n- `config` - Either:\n  - A keyword list with adapter configuration\n  - A Cloud module atom that implements the Buckets.Cloud behaviour","ref":"Buckets.Location.html#new/2-parameters"},{"type":"function","title":"Examples - Buckets.Location.new/2","doc":"# With explicit configuration\n    location = Location.new(\"uploads/file.pdf\", [\n      adapter: Buckets.Adapters.S3,\n      bucket: \"my-bucket\",\n      region: \"us-east-1\"\n    ])\n\n    # With Cloud module reference\n    location = Location.new(\"uploads/file.pdf\", MyApp.Cloud)","ref":"Buckets.Location.html#new/2-examples"},{"type":"function","title":"Returns - Buckets.Location.new/2","doc":"A `%Buckets.Location{}` struct with the path and configuration set.","ref":"Buckets.Location.html#new/2-returns"},{"type":"type","title":"Buckets.Location.t/0","doc":"","ref":"Buckets.Location.html#t:t/0"},{"type":"module","title":"Buckets.Object","doc":"Represents a file object in the Buckets storage system.\n\nAn Object encapsulates file data, metadata, and location information. Objects\ncan exist in various states:","ref":"Buckets.Object.html"},{"type":"module","title":"Object States - Buckets.Object","doc":"1. **Local only** - Created from a file or upload, not yet stored remotely\n   - `stored?: false`\n   - `location: %Buckets.Location.NotConfigured{}`\n   - `data: {:file, path}` or `{:data, binary}`\n\n2. **Stored remotely** - Uploaded to cloud storage\n   - `stored?: true`\n   - `location: %Buckets.Location{}`\n   - `data: nil` (unless loaded)\n\n3. **Loaded** - Remote object with data fetched\n   - `stored?: true`\n   - `location: %Buckets.Location{}`\n   - `data: {:file, path}` or `{:data, binary}`","ref":"Buckets.Object.html#module-object-states"},{"type":"module","title":"Creating Objects - Buckets.Object","doc":"Objects can be created from:\n- Files on disk: `from_file/1`\n- Plug uploads: `from_upload/1` with `%Plug.Upload{}`\n- LiveView uploads: `from_upload/1` with `%Phoenix.LiveView.UploadEntry{}`\n- Manually: `new/3`","ref":"Buckets.Object.html#module-creating-objects"},{"type":"module","title":"Metadata - Buckets.Object","doc":"Objects store metadata including:\n- `:content_type` - MIME type of the file\n- `:content_size` - Size in bytes\n- Custom metadata can be added as needed","ref":"Buckets.Object.html#module-metadata"},{"type":"module","title":"Example Usage - Buckets.Object","doc":"# Create from file\n    object = Buckets.Object.from_file(\"/path/to/document.pdf\")\n\n    # Upload to storage\n    {:ok, stored} = MyApp.Cloud.insert(object)\n\n    # Read data from stored object\n    {:ok, data} = MyApp.Cloud.read(stored)\n\n    # Load remote object data\n    {:ok, loaded} = MyApp.Cloud.load(stored)\n    {:ok, data} = Buckets.Object.read(loaded)","ref":"Buckets.Object.html#module-example-usage"},{"type":"function","title":"Buckets.Object.from_file/1","doc":"Creates an Object from a file on disk.\n\nAutomatically generates a UUID and extracts metadata including content type\n(from file extension) and content size.","ref":"Buckets.Object.html#from_file/1"},{"type":"function","title":"Parameters - Buckets.Object.from_file/1","doc":"- `path` - Path to the file on disk","ref":"Buckets.Object.html#from_file/1-parameters"},{"type":"function","title":"Raises - Buckets.Object.from_file/1","doc":"Raises if the file does not exist.","ref":"Buckets.Object.html#from_file/1-raises"},{"type":"function","title":"Examples - Buckets.Object.from_file/1","doc":"object = Buckets.Object.from_file(\"/tmp/upload.pdf\")\n    # %Buckets.Object{\n    #   uuid: \"generated-uuid\",\n    #   filename: \"upload.pdf\",\n    #   data: {:file, \"/tmp/upload.pdf\"},\n    #   metadata: %{\n    #     content_type: \"application/pdf\",\n    #     content_size: 12345\n    #   },\n    #   stored?: false\n    # }","ref":"Buckets.Object.html#from_file/1-examples"},{"type":"function","title":"Buckets.Object.from_upload/1","doc":"Creates an Object from various upload types.\n\nThis function has multiple clauses to handle different upload scenarios:","ref":"Buckets.Object.html#from_upload/1"},{"type":"function","title":"Plug.Upload - Buckets.Object.from_upload/1","doc":"Creates an Object from a `%Plug.Upload{}` struct (Phoenix controller uploads).\nAutomatically extracts metadata including content type and size.\n\n    def create(conn, %{\"file\" => upload}) do\n      object = Buckets.Object.from_upload(upload)\n      {:ok, stored} = MyApp.Cloud.insert(object)\n    end","ref":"Buckets.Object.html#from_upload/1-plug-upload"},{"type":"function","title":"LiveView UploadEntry (incomplete) - Buckets.Object.from_upload/1","doc":"Creates an Object from a `%Phoenix.LiveView.UploadEntry{}` that is not yet complete.\nUsed for generating signed URLs for direct uploads.\n\n    # In LiveView for generating upload URLs\n    object = Buckets.Object.from_upload(entry)\n    {:ok, config} = MyApp.Cloud.live_upload(entry)\n\nRaises if called with a completed upload entry without metadata.","ref":"Buckets.Object.html#from_upload/1-liveview-uploadentry-incomplete"},{"type":"function","title":"LiveView UploadEntry (completed) - Buckets.Object.from_upload/1","doc":"Creates an Object from a completed upload with metadata tuple `{entry, meta}`.\nThe metadata typically includes the uploaded file path and/or signed URL.\n\n    # After LiveView upload completes\n    consume_uploaded_entries(socket, :avatar, fn %{path: path}, entry ->\n      object = Buckets.Object.from_upload({entry, %{path: path}})\n      {:ok, object}\n    end)\n\nRaises if called with an incomplete upload entry.","ref":"Buckets.Object.html#from_upload/1-liveview-uploadentry-completed"},{"type":"function","title":"Buckets.Object.new/3","doc":"Creates a new Object with the given UUID, filename, and options.","ref":"Buckets.Object.html#new/3"},{"type":"function","title":"Parameters - Buckets.Object.new/3","doc":"- `uuid` - Unique identifier for the object\n- `filename` - Original filename\n- `opts` - Keyword list of options:\n  - `:location` - Either `{path, config}` tuple or `%Buckets.Location{}` struct\n  - `:metadata` - Map of metadata (defaults to `%{}`)","ref":"Buckets.Object.html#new/3-parameters"},{"type":"function","title":"Examples - Buckets.Object.new/3","doc":"# Basic object\n    object = Buckets.Object.new(\"123\", \"document.pdf\", [])\n\n    # With location\n    object = Buckets.Object.new(\"123\", \"document.pdf\",\n      location: {\"uploads/123/document.pdf\", MyApp.Cloud}\n    )\n\n    # With metadata\n    object = Buckets.Object.new(\"123\", \"document.pdf\",\n      metadata: %{content_type: \"application/pdf\", author: \"John\"}\n    )","ref":"Buckets.Object.html#new/3-examples"},{"type":"function","title":"Buckets.Object.read/1","doc":"Reads the data from an object.\n\nReturns `{:ok, data}` if the object has data loaded, either in memory\nor from a file. Returns `{:error, :not_loaded}` if no data is available.","ref":"Buckets.Object.html#read/1"},{"type":"function","title":"Examples - Buckets.Object.read/1","doc":"# Object with data in memory\n    {:ok, binary_data} = Buckets.Object.read(object)\n\n    # Object without loaded data\n    {:error, :not_loaded} = Buckets.Object.read(remote_object)","ref":"Buckets.Object.html#read/1-examples"},{"type":"function","title":"Note - Buckets.Object.read/1","doc":"For remote objects without loaded data, use `Cloud.load/2` first to fetch\nthe data from storage.","ref":"Buckets.Object.html#read/1-note"},{"type":"function","title":"Buckets.Object.read!/1","doc":"Reads the data from an object, raising on error.\n\nSame as `read/1` but raises an exception if the data cannot be read.","ref":"Buckets.Object.html#read!/1"},{"type":"function","title":"Raises - Buckets.Object.read!/1","doc":"- Raises if the object has no data loaded\n- Raises if the file cannot be read from disk","ref":"Buckets.Object.html#read!/1-raises"},{"type":"function","title":"Examples - Buckets.Object.read!/1","doc":"# Success case\n    binary_data = Buckets.Object.read!(object)\n\n    # Raises for unloaded object\n    Buckets.Object.read!(remote_object)\n    # ** (RuntimeError) Called `read!/1` with object that does not have data loaded.","ref":"Buckets.Object.html#read!/1-examples"},{"type":"type","title":"Buckets.Object.t/0","doc":"","ref":"Buckets.Object.html#t:t/0"},{"type":"module","title":"Buckets.SignedURL","doc":"Represents a signed URL for direct access to objects in cloud storage.\n\nSigned URLs provide temporary, secure access to objects without requiring\nauthentication. They are commonly used for:\n- Direct file downloads from cloud storage\n- Direct uploads to cloud storage (especially with LiveView)\n- Temporary sharing of private files","ref":"Buckets.SignedURL.html"},{"type":"module","title":"Structure - Buckets.SignedURL","doc":"A SignedURL contains:\n- `:url` - The complete signed URL string that can be used for access\n- `:location` - The `Buckets.Location` struct indicating where the object is stored","ref":"Buckets.SignedURL.html#module-structure"},{"type":"module","title":"Usage - Buckets.SignedURL","doc":"SignedURLs are typically generated by adapters through the `url/2` callback:\n\n    {:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 3600)\n\n    # Access the URL string\n    url_string = signed_url.url\n    # or use String.Chars protocol\n    url_string = to_string(signed_url)","ref":"Buckets.SignedURL.html#module-usage"},{"type":"module","title":"LiveView Uploads - Buckets.SignedURL","doc":"SignedURLs are used for direct uploads in LiveView:\n\n    {:ok, config} = MyApp.Cloud.live_upload(entry)\n    # config.url is a SignedURL for uploading","ref":"Buckets.SignedURL.html#module-liveview-uploads"},{"type":"module","title":"JSON Encoding - Buckets.SignedURL","doc":"SignedURLs implement the `Jason.Encoder` protocol and encode to just the URL string:\n\n    Jason.encode!(%{upload_url: signed_url})\n    # {\"upload_url\": \"https://storage.example.com/bucket/file?signature=...\"}","ref":"Buckets.SignedURL.html#module-json-encoding"},{"type":"module","title":"Security Considerations - Buckets.SignedURL","doc":"- Signed URLs typically expire after a set time period\n- They should be treated as sensitive data\n- Different adapters have different expiration limits and security features","ref":"Buckets.SignedURL.html#module-security-considerations"},{"type":"function","title":"Buckets.SignedURL.to_string/1","doc":"Converts a SignedURL to its URL string representation.\n\nThis function extracts just the URL string from the SignedURL struct.\nIt's also used by the `String.Chars` protocol implementation.","ref":"Buckets.SignedURL.html#to_string/1"},{"type":"function","title":"Examples - Buckets.SignedURL.to_string/1","doc":"iex> signed_url = %Buckets.SignedURL{url: \"https://example.com/file?sig=123\"}\n    iex> Buckets.SignedURL.to_string(signed_url)\n    \"https://example.com/file?sig=123\"\n\n    # Also works with Kernel.to_string/1\n    iex> to_string(signed_url)\n    \"https://example.com/file?sig=123\"","ref":"Buckets.SignedURL.html#to_string/1-examples"},{"type":"type","title":"Buckets.SignedURL.t/0","doc":"","ref":"Buckets.SignedURL.html#t:t/0"},{"type":"module","title":"Buckets.Adapters.GCS","doc":"Google Cloud Storage adapter for Buckets.\n\nThis adapter provides a native implementation for GCS operations using\nonly `:req` as a dependency.","ref":"Buckets.Adapters.GCS.html"},{"type":"module","title":"Setup - Buckets.Adapters.GCS","doc":"To use this adapter, you need to start your Cloud module in your application's\nsupervision tree to enable automatic token caching and refresh:\n\n    children = [\n      # ... your other processes\n      MyApp.Cloud\n    ]\n\n    Supervisor.start_link(children, opts)\n\nThe Cloud module will automatically start the required authentication processes for GCS locations.\n\nThe supervisor will automatically manage authentication tokens for each unique set of\nservice account credentials, refreshing them before they expire.","ref":"Buckets.Adapters.GCS.html#module-setup"},{"type":"module","title":"Configuration - Buckets.Adapters.GCS","doc":"You can configure GCS using either service account credentials as a JSON string\nor by providing a path to a service account JSON file:\n\n    # Using credentials from environment variable (JSON string)\n    config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.GCS,\n      bucket: \"my-bucket\",\n      service_account_credentials: System.fetch_env!(\"GOOGLE_CREDENTIALS\")\n\n    # Using credentials from file path\n    config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.GCS,\n      bucket: \"my-bucket\",\n      service_account_path: \"/path/to/service-account.json\"","ref":"Buckets.Adapters.GCS.html#module-configuration"},{"type":"module","title":"Optional Configuration - Buckets.Adapters.GCS","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.GCS,\n      bucket: \"my-bucket\",\n      service_account_credentials: System.fetch_env!(\"GOOGLE_CREDENTIALS\"),\n      # Optional: base path within bucket\n      path: \"uploads\",\n      # Optional: uploader module for LiveView direct uploads\n      uploader: \"GCS\"\n\nThe `:service_account_credentials` option accepts a JSON string, making it easy to pass\ncredentials via environment variables in production environments.","ref":"Buckets.Adapters.GCS.html#module-optional-configuration"},{"type":"module","title":"Performance - Buckets.Adapters.GCS","doc":"This implementation automatically caches and refreshes Google Cloud access tokens,\nsignificantly reducing latency compared to generating new tokens for each request.\nTokens are refreshed 5 minutes before expiration to ensure uninterrupted service.","ref":"Buckets.Adapters.GCS.html#module-performance"},{"type":"function","title":"Buckets.Adapters.GCS.url/2","doc":"Gets a signed URL for temporarily delegating access to an object in a bucket.","ref":"Buckets.Adapters.GCS.html#url/2"},{"type":"function","title":"Errors - Buckets.Adapters.GCS.url/2","doc":"* 403 PERMISSION_DENIED: Permission 'iam.serviceAccounts.signBlob' denied on resource\n      (or it may not exist). Make sure the authorized SA has role roles/iam.serviceAccountTokenCreator\n      on the SA passed in the URL.\n\n    * 403 PERMISSION_DENIED: IAM Service Account Credentials API has not been used in project {project_id}\n      before or it is disabled. Enable it by visiting {url} then retry. If you enabled this API recently,\n      wait a few minutes for the action to propagate to our systems and retry. Make sure the authorized SA\n      has role roles/iam.serviceAccountTokenCreator on the SA passed in the URL.\"","ref":"Buckets.Adapters.GCS.html#url/2-errors"},{"type":"module","title":"Buckets.Adapters.S3","doc":"Adapter for Amazon S3 and S3-compatible storage services.\n\nThis adapter supports multiple S3-compatible providers through the `:provider` option:\n- `:aws` (default) - Amazon S3\n- `:cloudflare` - Cloudflare R2\n- `:digitalocean` - DigitalOcean Spaces\n- `:tigris` - Tigris","ref":"Buckets.Adapters.S3.html"},{"type":"module","title":"Configuration - Buckets.Adapters.S3","doc":"","ref":"Buckets.Adapters.S3.html#module-configuration"},{"type":"module","title":"Amazon S3 (default) - Buckets.Adapters.S3","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.S3,\n      bucket: \"my-bucket\",\n      region: \"us-east-1\",\n      access_key_id: System.fetch_env!(\"AWS_ACCESS_KEY_ID\"),\n      secret_access_key: System.fetch_env!(\"AWS_SECRET_ACCESS_KEY\")","ref":"Buckets.Adapters.S3.html#module-amazon-s3-default"},{"type":"module","title":"Cloudflare R2 - Buckets.Adapters.S3","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.S3,\n      provider: :cloudflare,\n      bucket: \"my-bucket\",\n      endpoint_url: \"https://YOUR_ACCOUNT_ID.r2.cloudflarestorage.com\",\n      access_key_id: System.fetch_env!(\"R2_ACCESS_KEY_ID\"),\n      secret_access_key: System.fetch_env!(\"R2_SECRET_ACCESS_KEY\")\n\nNote: Cloudflare R2 automatically uses `region: \"auto\"`.","ref":"Buckets.Adapters.S3.html#module-cloudflare-r2"},{"type":"module","title":"DigitalOcean Spaces - Buckets.Adapters.S3","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.S3,\n      provider: :digitalocean,\n      bucket: \"my-bucket\",\n      access_key_id: System.fetch_env!(\"SPACES_ACCESS_KEY\"),\n      secret_access_key: System.fetch_env!(\"SPACES_SECRET_KEY\")\n\nNote: DigitalOcean Spaces defaults to NYC3 region and endpoint. Override with:\n\n    config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.S3,\n      provider: :digitalocean,\n      bucket: \"my-bucket\",\n      region: \"sfo3\",\n      endpoint_url: \"https://sfo3.digitaloceanspaces.com\",\n      access_key_id: System.fetch_env!(\"SPACES_ACCESS_KEY\"),\n      secret_access_key: System.fetch_env!(\"SPACES_SECRET_KEY\")","ref":"Buckets.Adapters.S3.html#module-digitalocean-spaces"},{"type":"module","title":"Tigris - Buckets.Adapters.S3","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.S3,\n      provider: :tigris,\n      bucket: \"my-bucket\",\n      region: \"auto\",\n      access_key_id: System.fetch_env!(\"TIGRIS_ACCESS_KEY_ID\"),\n      secret_access_key: System.fetch_env!(\"TIGRIS_SECRET_ACCESS_KEY\")\n\nNote: Tigris automatically uses the endpoint `https://fly.storage.tigris.dev`.","ref":"Buckets.Adapters.S3.html#module-tigris"},{"type":"module","title":"Required Options - Buckets.Adapters.S3","doc":"- `:bucket` - The S3 bucket name\n- `:access_key_id` - AWS access key ID or equivalent\n- `:secret_access_key` - AWS secret access key or equivalent\n- `:region` - AWS region (defaults to \"auto\" for some providers)\n- `:endpoint_url` - Required for non-AWS providers","ref":"Buckets.Adapters.S3.html#module-required-options"},{"type":"module","title":"Optional Options - Buckets.Adapters.S3","doc":"- `:provider` - The S3-compatible provider (defaults to `:aws`)\n- `:path` - Base path within the bucket\n- `:uploader` - Uploader configuration for direct uploads","ref":"Buckets.Adapters.S3.html#module-optional-options"},{"type":"module","title":"Supervision - Buckets.Adapters.S3","doc":"This adapter does not require any supervised processes. You do not need to add Cloud modules\nusing this adapter to your supervision tree.","ref":"Buckets.Adapters.S3.html#module-supervision"},{"type":"module","title":"Buckets.Adapters.Volume","doc":"Adapter for local filesystem storage.\n\nThe Volume adapter stores files on the local filesystem and is primarily intended\nfor development environments. It supports signed URLs and integrates with the\nBuckets dev router for handling uploads and downloads.","ref":"Buckets.Adapters.Volume.html"},{"type":"module","title":"Configuration - Buckets.Adapters.Volume","doc":"config :my_app, MyApp.Cloud,\n      adapter: Buckets.Adapters.Volume,\n      bucket: \"tmp/buckets_volume\",\n      base_url: \"http://localhost:4000\"","ref":"Buckets.Adapters.Volume.html#module-configuration"},{"type":"module","title":"Required Options - Buckets.Adapters.Volume","doc":"- `:bucket` - The local directory path where files will be stored\n- `:base_url` - The base URL for generating file URLs (e.g., your Phoenix app URL)","ref":"Buckets.Adapters.Volume.html#module-required-options"},{"type":"module","title":"Optional Options - Buckets.Adapters.Volume","doc":"- `:path` - Base path within the bucket directory\n- `:endpoint` - Custom endpoint path for signed URLs (defaults to \"/__buckets__\")\n- `:uploader` - Uploader configuration for direct uploads","ref":"Buckets.Adapters.Volume.html#module-optional-options"},{"type":"module","title":"Development Router Integration - Buckets.Adapters.Volume","doc":"To handle local uploads and downloads in development, add the Volume routes to your router:\n\n    # In your router.ex\n    if Application.compile_env(:my_app, :dev_routes) do\n      import Buckets.Router\n\n      buckets_volume(MyApp.Cloud)\n    end\n\nThis will mount routes at `/__buckets__/:bucket/*path` for handling:\n- File uploads (PUT requests)\n- File downloads (GET requests)","ref":"Buckets.Adapters.Volume.html#module-development-router-integration"},{"type":"module","title":"Signed URLs - Buckets.Adapters.Volume","doc":"The Volume adapter supports signed URLs for secure file uploads and downloads:\n\n    {:ok, object} = MyApp.Cloud.url(object, expires_in: 3600)\n    # Returns object with signed URL valid for 1 hour","ref":"Buckets.Adapters.Volume.html#module-signed-urls"},{"type":"module","title":"Supervision - Buckets.Adapters.Volume","doc":"This adapter does not require any supervised processes. Do not add Cloud modules\nusing this adapter to your supervision tree.","ref":"Buckets.Adapters.Volume.html#module-supervision"},{"type":"function","title":"Buckets.Adapters.Volume.verify_signed_path/3","doc":"","ref":"Buckets.Adapters.Volume.html#verify_signed_path/3"},{"type":"behaviour","title":"Buckets.Adapter","doc":"Behaviour definition for Buckets storage adapters.\n\nAn adapter implements the storage operations for a specific backend (filesystem,\nS3, GCS, etc.). All adapters must implement the required callbacks to handle\nfile operations like uploading, downloading, and generating signed URLs.","ref":"Buckets.Adapter.html"},{"type":"behaviour","title":"Implementing an Adapter - Buckets.Adapter","doc":"To create a new adapter, implement this behaviour:\n\n    defmodule MyApp.CustomAdapter do\n      @behaviour Buckets.Adapter\n\n      @impl true\n      def validate_config(config) do\n        # Validate and normalize configuration\n      end\n\n      @impl true\n      def put(object, remote_path, config) do\n        # Upload object to storage\n      end\n\n      @impl true\n      def get(remote_path, config) do\n        # Download data from storage\n      end\n\n      @impl true\n      def url(remote_path, config) do\n        # Generate signed URL for direct access\n      end\n\n      @impl true\n      def delete(remote_path, config) do\n        # Delete object from storage\n      end\n\n      # Optional: Only if adapter needs supervised processes\n      @impl true\n      def child_spec(config, cloud_module) do\n        # Return supervisor child specification\n      end\n    end","ref":"Buckets.Adapter.html#module-implementing-an-adapter"},{"type":"behaviour","title":"Configuration - Buckets.Adapter","doc":"Each adapter defines its own configuration options, but common ones include:\n- `:bucket` - The storage bucket/container name\n- `:path` - Base path within the bucket\n- `:uploader` - Uploader type for LiveView direct uploads","ref":"Buckets.Adapter.html#module-configuration"},{"type":"behaviour","title":"Supervised Processes - Buckets.Adapter","doc":"Some adapters (like GCS) require background processes for authentication or\nconnection management. These adapters should implement the optional `child_spec/2`\ncallback. Adapters that don't need supervision (like Volume, S3) should not\nimplement this callback.","ref":"Buckets.Adapter.html#module-supervised-processes"},{"type":"callback","title":"Buckets.Adapter.child_spec/2","doc":"Returns a child specification for processes this adapter needs, if any.\n\nThis callback is optional. Adapters that need supervised processes (like GCS\nauth servers) should implement this callback. Adapters that don't need any\nsupervised processes (like Volume, S3) can omit this callback entirely.\n\nWhen implemented, should return:\n- `child_spec()` - A supervisor child specification\n- `{:error, reason}` - Config is invalid\n\nExamples:\n- GCS adapter implements this to return auth server child spec\n- Volume/S3 adapters don't implement this callback at all","ref":"Buckets.Adapter.html#c:child_spec/2"},{"type":"callback","title":"Buckets.Adapter.delete/2","doc":"Deletes an object from the storage backend.\n\nTakes the remote path of the object to delete and the adapter configuration.\nReturns `{:ok, metadata}` on success or `{:error, reason}` on failure.","ref":"Buckets.Adapter.html#c:delete/2"},{"type":"callback","title":"Examples - Buckets.Adapter.delete/2","doc":"def delete(\"uploads/file.pdf\", config) do\n      case delete_object(path, config) do\n        :ok -> {:ok, %{}}\n        {:error, reason} -> {:error, reason}\n      end\n    end","ref":"Buckets.Adapter.html#c:delete/2-examples"},{"type":"callback","title":"Buckets.Adapter.get/2","doc":"Downloads data from the storage backend.\n\nTakes the remote path of the object and the adapter configuration.\nReturns the binary data on success or an error tuple on failure.","ref":"Buckets.Adapter.html#c:get/2"},{"type":"callback","title":"Examples - Buckets.Adapter.get/2","doc":"def get(\"uploads/file.pdf\", config) do\n      case download_file(path, config) do\n        {:ok, data} -> {:ok, data}\n        {:error, :not_found} -> {:error, :not_found}\n        {:error, reason} -> {:error, reason}\n      end\n    end","ref":"Buckets.Adapter.html#c:get/2-examples"},{"type":"callback","title":"Buckets.Adapter.put/3","doc":"Uploads an object to the storage backend.\n\nTakes a `Buckets.Object` struct containing the data to upload, the remote\npath where it should be stored, and the adapter configuration.\n\nShould return `{:ok, metadata}` on success where metadata is adapter-specific,\nor `{:error, reason}` on failure.","ref":"Buckets.Adapter.html#c:put/3"},{"type":"callback","title":"Examples - Buckets.Adapter.put/3","doc":"def put(object, \"uploads/file.pdf\", config) do\n      # Upload logic here\n      {:ok, %{etag: \"abc123\", version: \"v1\"}}\n    end","ref":"Buckets.Adapter.html#c:put/3-examples"},{"type":"callback","title":"Buckets.Adapter.url/2","doc":"Generates a signed URL for direct access to an object.\n\nTakes the remote path and configuration (which may include options like\n`:expires_in` for controlling URL expiration).\n\nReturns a `Buckets.SignedURL` struct containing the URL and metadata.","ref":"Buckets.Adapter.html#c:url/2"},{"type":"callback","title":"Options - Buckets.Adapter.url/2","doc":"- `:expires_in` - URL expiration time in seconds (adapter-specific default)\n- `:for_upload` - Generate a URL for uploading instead of downloading","ref":"Buckets.Adapter.html#c:url/2-options"},{"type":"callback","title":"Examples - Buckets.Adapter.url/2","doc":"def url(\"uploads/file.pdf\", config) do\n      signed_url = generate_signed_url(path, config)\n      {:ok, %Buckets.SignedURL{url: signed_url, expires_at: expiry}}\n    end","ref":"Buckets.Adapter.html#c:url/2-examples"},{"type":"callback","title":"Buckets.Adapter.validate_config/1","doc":"Validates and normalizes adapter configuration.\n\nThis callback is called during Cloud module initialization to ensure the\nconfiguration is valid. It should:\n- Validate required options are present\n- Set default values for optional options\n- Normalize configuration format\n- Return an error tuple with invalid/missing keys if validation fails","ref":"Buckets.Adapter.html#c:validate_config/1"},{"type":"callback","title":"Examples - Buckets.Adapter.validate_config/1","doc":"def validate_config(config) do\n      config\n      |> Keyword.validate([:adapter, :bucket, :path, :uploader])\n      |> case do\n        {:ok, config} -> validate_required(config, [:bucket])\n        error -> error\n      end\n    end","ref":"Buckets.Adapter.html#c:validate_config/1-examples"},{"type":"function","title":"Buckets.Adapter.validate_required/2","doc":"Validates that required configuration keys are present.\n\nA helper function for adapters to use in their `validate_config/1` callback\nto ensure all required configuration options are provided.","ref":"Buckets.Adapter.html#validate_required/2"},{"type":"function","title":"Parameters - Buckets.Adapter.validate_required/2","doc":"- `config` - The configuration keyword list to validate\n- `required` - List of required keys that must be present in the config","ref":"Buckets.Adapter.html#validate_required/2-parameters"},{"type":"function","title":"Returns - Buckets.Adapter.validate_required/2","doc":"- `{:ok, config}` - All required keys are present\n- `{:error, missing_keys}` - List of keys that are missing from config","ref":"Buckets.Adapter.html#validate_required/2-returns"},{"type":"function","title":"Examples - Buckets.Adapter.validate_required/2","doc":"def validate_config(config) do\n      with {:ok, config} <- Keyword.validate(config, @allowed_options),\n           {:ok, config} <- validate_required(config, [:bucket, :access_key]) do\n        {:ok, config}\n      end\n    end","ref":"Buckets.Adapter.html#validate_required/2-examples"},{"type":"function","title":"Implementation Note - Buckets.Adapter.validate_required/2","doc":"The function filters out `nil` values from the required list, allowing\nconditional validation based on other configuration options.","ref":"Buckets.Adapter.html#validate_required/2-implementation-note"},{"type":"module","title":"Buckets.Router","doc":"Router helpers for integrating Buckets with Phoenix applications.\n\nThis module provides macros for setting up development routes that handle\nfile uploads and downloads when using the Volume adapter. These routes\nenable direct file uploads to the local filesystem during development,\nmimicking the behavior of cloud storage providers.","ref":"Buckets.Router.html"},{"type":"module","title":"Usage - Buckets.Router","doc":"Import this module in your Phoenix router and use the `buckets_volume/2`\nmacro to set up the necessary routes:\n\n    # In your router.ex\n    defmodule MyAppWeb.Router do\n      use MyAppWeb, :router\n\n      # Only include in development\n      if Application.compile_env(:my_app, :dev_routes) do\n        import Buckets.Router\n\n        # Sets up routes at /__buckets__/:bucket/*path\n        buckets_volume(MyApp.Cloud)\n      end\n    end","ref":"Buckets.Router.html#module-usage"},{"type":"module","title":"Routes Created - Buckets.Router","doc":"The macro creates two routes under the `/__buckets__` scope:\n- `GET /__buckets__/:bucket/*path` - Download files\n- `PUT /__buckets__/:bucket/*path` - Upload files","ref":"Buckets.Router.html#module-routes-created"},{"type":"module","title":"Security - Buckets.Router","doc":"These routes validate:\n- The bucket name matches the configured bucket\n- Upload requests have valid signatures (generated by the Volume adapter)","ref":"Buckets.Router.html#module-security"},{"type":"module","title":"Configuration - Buckets.Router","doc":"The routes can be customized with options:\n\n    buckets_volume(MyApp.Cloud, cache_control: 3600)\n\nAvailable options are passed to the controller and may include:\n- `:cache_control` - Max age in seconds for cache control headers","ref":"Buckets.Router.html#module-configuration"},{"type":"macro","title":"Buckets.Router.buckets_volume/2","doc":"Sets up Volume adapter routes for local file handling in development.\n\nThis macro creates GET and PUT routes under the `/__buckets__` scope that\nhandle file downloads and uploads respectively. It's designed to be used\nwith the Volume adapter during development to provide a local alternative\nto cloud storage.","ref":"Buckets.Router.html#buckets_volume/2"},{"type":"macro","title":"Parameters - Buckets.Router.buckets_volume/2","doc":"- `cloud_module` - The Cloud module to use for configuration (e.g., `MyApp.Cloud`)\n- `opts` - Optional keyword list of options:\n  - `:cache_control` - Max age in seconds for cache control headers","ref":"Buckets.Router.html#buckets_volume/2-parameters"},{"type":"macro","title":"Examples - Buckets.Router.buckets_volume/2","doc":"# Basic usage\n    buckets_volume(MyApp.Cloud)\n\n    # With cache control\n    buckets_volume(MyApp.Cloud, cache_control: 3600)","ref":"Buckets.Router.html#buckets_volume/2-examples"},{"type":"macro","title":"Routes Created - Buckets.Router.buckets_volume/2","doc":"The following routes are created:\n- `GET /__buckets__/:bucket/*path` - Downloads a file from the local filesystem\n- `PUT /__buckets__/:bucket/*path` - Uploads a file to the local filesystem\n\nBoth routes include the cloud module and options in the connection's private\ndata for use by the VolumeController.","ref":"Buckets.Router.html#buckets_volume/2-routes-created"},{"type":"macro","title":"Security Note - Buckets.Router.buckets_volume/2","doc":"These routes should only be used in development. The macro includes validation\nto ensure the bucket parameter matches the configured bucket and that uploads\ninclude valid signatures.","ref":"Buckets.Router.html#buckets_volume/2-security-note"},{"type":"function","title":"Buckets.Router.scope/0","doc":"Returns the default scope path used for Buckets routes.\n\nThis is useful if you need to reference the scope in your own code or\nneed to know where the routes are mounted.","ref":"Buckets.Router.html#scope/0"},{"type":"function","title":"Examples - Buckets.Router.scope/0","doc":"iex> Buckets.Router.scope()\n    \"__buckets__\"","ref":"Buckets.Router.html#scope/0-examples"},{"type":"module","title":"Buckets.Router.VolumeController","doc":"Phoenix controller for handling local file uploads and downloads in development.\n\nThis controller is automatically used by the `Buckets.Router.buckets_volume/2`\nmacro to handle Volume adapter file operations.","ref":"Buckets.Router.VolumeController.html"},{"type":"module","title":"Endpoints - Buckets.Router.VolumeController","doc":"- `get/2` - Downloads files from the local filesystem\n- `put/2` - Uploads files to the local filesystem with signature validation\n\nFor usage and configuration, see `Buckets.Router`.","ref":"Buckets.Router.VolumeController.html#module-endpoints"},{"type":"function","title":"Buckets.Router.VolumeController.get/2","doc":"","ref":"Buckets.Router.VolumeController.html#get/2"},{"type":"function","title":"Buckets.Router.VolumeController.put/2","doc":"","ref":"Buckets.Router.VolumeController.html#put/2"},{"type":"module","title":"Buckets.Telemetry","doc":"Telemetry events for Buckets.\n\nThis module defines telemetry events that other applications can attach to.","ref":"Buckets.Telemetry.html"},{"type":"module","title":"Event Format - Buckets.Telemetry","doc":"Events follow the convention: `[:buckets, component, operation, stage]`","ref":"Buckets.Telemetry.html#module-event-format"},{"type":"module","title":"Events - Buckets.Telemetry","doc":"","ref":"Buckets.Telemetry.html#module-events"},{"type":"module","title":"Cloud Operations - Buckets.Telemetry","doc":"* `[:buckets, :cloud, :insert, :start]` - Emitted when a file upload operation starts\n* `[:buckets, :cloud, :insert, :stop]` - Emitted when a file upload operation completes\n* `[:buckets, :cloud, :insert, :exception]` - Emitted when a file upload operation raises an exception\n* `[:buckets, :cloud, :delete, :start]` - Emitted when a file deletion operation starts\n* `[:buckets, :cloud, :delete, :stop]` - Emitted when a file deletion operation completes\n* `[:buckets, :cloud, :delete, :exception]` - Emitted when a file deletion operation raises an exception\n* `[:buckets, :cloud, :read, :start]` - Emitted when a file read operation starts\n* `[:buckets, :cloud, :read, :stop]` - Emitted when a file read operation completes\n* `[:buckets, :cloud, :read, :exception]` - Emitted when a file read operation raises an exception\n* `[:buckets, :cloud, :load, :start]` - Emitted when a file load operation starts\n* `[:buckets, :cloud, :load, :stop]` - Emitted when a file load operation completes\n* `[:buckets, :cloud, :load, :exception]` - Emitted when a file load operation raises an exception\n* `[:buckets, :cloud, :url, :start]` - Emitted when a URL generation operation starts\n* `[:buckets, :cloud, :url, :stop]` - Emitted when a URL generation operation completes\n* `[:buckets, :cloud, :url, :exception]` - Emitted when a URL generation operation raises an exception","ref":"Buckets.Telemetry.html#module-cloud-operations"},{"type":"module","title":"Adapter Operations - Buckets.Telemetry","doc":"* `[:buckets, :adapter, :put, :start]` - Emitted when an adapter put operation starts\n* `[:buckets, :adapter, :put, :stop]` - Emitted when an adapter put operation completes\n* `[:buckets, :adapter, :put, :exception]` - Emitted when an adapter put operation raises an exception\n* `[:buckets, :adapter, :get, :start]` - Emitted when an adapter get operation starts\n* `[:buckets, :adapter, :get, :stop]` - Emitted when an adapter get operation completes\n* `[:buckets, :adapter, :get, :exception]` - Emitted when an adapter get operation raises an exception\n* `[:buckets, :adapter, :url, :start]` - Emitted when an adapter URL generation starts\n* `[:buckets, :adapter, :url, :stop]` - Emitted when an adapter URL generation completes\n* `[:buckets, :adapter, :url, :exception]` - Emitted when an adapter URL generation raises an exception\n* `[:buckets, :adapter, :delete, :start]` - Emitted when an adapter delete operation starts\n* `[:buckets, :adapter, :delete, :stop]` - Emitted when an adapter delete operation completes\n* `[:buckets, :adapter, :delete, :exception]` - Emitted when an adapter delete operation raises an exception","ref":"Buckets.Telemetry.html#module-adapter-operations"},{"type":"module","title":"Authentication Operations - Buckets.Telemetry","doc":"* `[:buckets, :auth, :token, :fetch, :start]` - Emitted when token fetching starts\n* `[:buckets, :auth, :token, :fetch, :stop]` - Emitted when token fetching completes\n* `[:buckets, :auth, :token, :fetch, :exception]` - Emitted when token fetching raises an exception\n* `[:buckets, :auth, :token, :refresh, :start]` - Emitted when token refresh starts\n* `[:buckets, :auth, :token, :refresh, :stop]` - Emitted when token refresh completes\n* `[:buckets, :auth, :token, :refresh, :exception]` - Emitted when token refresh raises an exception","ref":"Buckets.Telemetry.html#module-authentication-operations"},{"type":"module","title":"Usage - Buckets.Telemetry","doc":"To attach to these events in your application:\n\n```elixir\n:telemetry.attach(\n  \"my-handler-id\",\n  [:buckets, :cloud, :insert, :start],\n  &MyApp.handle_cloud_insert/4,\n  nil\n)\n\ndef handle_cloud_insert(_event_name, measurements, metadata, _config) do\n  # Process the event\n  IO.inspect(measurements)\n  IO.inspect(metadata)\nend\n```\n\nYou may attach a default logger handler, if you want basic logs for all telemetry events:\n\n```elixir\nBuckets.Telemetry.attach_default_logger(:debug)\n```","ref":"Buckets.Telemetry.html#module-usage"},{"type":"function","title":"Buckets.Telemetry.attach_default_logger/1","doc":"Attaches a default structured JSON Telemetry handler for logging.","ref":"Buckets.Telemetry.html#attach_default_logger/1"},{"type":"function","title":"Buckets.Telemetry.default_handler_id/0","doc":"","ref":"Buckets.Telemetry.html#default_handler_id/0"},{"type":"function","title":"Buckets.Telemetry.detach_default_logger/0","doc":"Undoes `Buckets.Telemetry.attach_default_logger/1` by detaching the attached logger.","ref":"Buckets.Telemetry.html#detach_default_logger/0"},{"type":"function","title":"Examples - Buckets.Telemetry.detach_default_logger/0","doc":"Detach a previously attached logger:\n\n    :ok = Buckets.Telemetry.attach_default_logger()\n    :ok = Buckets.Telemetry.detach_default_logger()\n\nAttempt to detach when a logger wasn't attached:\n\n    {:error, :not_found} = Buckets.Telemetry.detach_default_logger()","ref":"Buckets.Telemetry.html#detach_default_logger/0-examples"},{"type":"function","title":"Buckets.Telemetry.emit_event/3","doc":"Emits a telemetry event with the given name, measurements, and metadata.","ref":"Buckets.Telemetry.html#emit_event/3"},{"type":"function","title":"Buckets.Telemetry.handle_event/4","doc":"Default telemetry event handler that logs events.","ref":"Buckets.Telemetry.html#handle_event/4"},{"type":"function","title":"Buckets.Telemetry.span/3","doc":"Wraps a function call with start and stop telemetry events.","ref":"Buckets.Telemetry.html#span/3"},{"type":"function","title":"Buckets.Telemetry.start_event/2","doc":"Emits a start event and returns a function to emit the corresponding stop event.\n\nThis is useful for span-like events where you want to measure the duration of an operation.","ref":"Buckets.Telemetry.html#start_event/2"},{"type":"module","title":"Buckets.Util","doc":"Utility functions for the Buckets library.\n\nThis module provides helper functions used internally by Buckets for\nfile operations and string manipulation.","ref":"Buckets.Util.html"},{"type":"function","title":"Buckets.Util.normalize_filename/1","doc":"Normalizes a filename for safe storage in cloud systems.\n\nThis function sanitizes filenames by:\n- Replacing whitespace with underscores\n- Removing all characters except letters, numbers, dots, underscores, and hyphens\n\nThis normalization helps prevent issues with special characters in different\nstorage systems and URLs.\n\nThis function is used by Cloud modules when generating storage paths. It can be overridden\nin your Cloud module if you need different normalization rules:\n\n    defmodule MyApp.Cloud do\n      use Buckets.Cloud, otp_app: :my_app\n\n      def normalize_filename(filename) do\n        # Custom normalization logic\n      end\n    end","ref":"Buckets.Util.html#normalize_filename/1"},{"type":"function","title":"Examples - Buckets.Util.normalize_filename/1","doc":"iex> Buckets.Util.normalize_filename(\"My Document.pdf\")\n    \"My_Document.pdf\"\n\n    iex> Buckets.Util.normalize_filename(\"invoice #123 (final).xlsx\")\n    \"invoice_123_final.xlsx\"\n\n    iex> Buckets.Util.normalize_filename(\"photo@event!.jpg\")\n    \"photoevent.jpg\"","ref":"Buckets.Util.html#normalize_filename/1-examples"},{"type":"function","title":"Buckets.Util.size/1","doc":"Gets the size of a file in bytes. Raises if the file does not exist or cannot be accessed","ref":"Buckets.Util.html#size/1"},{"type":"function","title":"Examples - Buckets.Util.size/1","doc":"iex> Buckets.Util.size(\"/path/to/file.pdf\")\n    12345\n\n    iex> Buckets.Util.size(\"non_existent.txt\")\n    ** (File.Error) could not read file stats \"non_existent.txt\": no such file or directory","ref":"Buckets.Util.html#size/1-examples"},{"type":"module","title":"Buckets.Adapters.GCS.Auth","doc":"Google Cloud Storage authentication using service account credentials.\n\nHandles JWT token generation and OAuth2 token exchange without external dependencies.","ref":"Buckets.Adapters.GCS.Auth.html"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.exchange_jwt_for_token/1","doc":"Exchanges a JWT token for an access token via Google's OAuth2 endpoint.","ref":"Buckets.Adapters.GCS.Auth.html#exchange_jwt_for_token/1"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.generate_jwt/1","doc":"Generates a JWT token for Google service account authentication.","ref":"Buckets.Adapters.GCS.Auth.html#generate_jwt/1"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.get_access_token/1","doc":"Gets an access token for GCS API access.","ref":"Buckets.Adapters.GCS.Auth.html#get_access_token/1"},{"type":"function","title":"Examples - Buckets.Adapters.GCS.Auth.get_access_token/1","doc":"iex> credentials = %{\n    ...>   \"client_email\" => \"test@example.com\",\n    ...>   \"private_key\" => \"-----BEGIN PRIVATE KEY-----\\n...\\n-----END PRIVATE KEY-----\"\n    ...> }\n    iex> Buckets.GCS.Auth.get_access_token(credentials)\n    {:ok, \"ya29.c.Ko8...\"}","ref":"Buckets.Adapters.GCS.Auth.html#get_access_token/1-examples"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.get_credentials/1","doc":"Get service account credentials from a location config.","ref":"Buckets.Adapters.GCS.Auth.html#get_credentials/1"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.load_credentials/1","doc":"Loads service account credentials from a file path.","ref":"Buckets.Adapters.GCS.Auth.html#load_credentials/1"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.parse_credentials/1","doc":"Parse service account credentials from a JSON string.","ref":"Buckets.Adapters.GCS.Auth.html#parse_credentials/1"},{"type":"function","title":"Buckets.Adapters.GCS.Auth.validate_credentials/1","doc":"Validates that service account credentials contain required fields.","ref":"Buckets.Adapters.GCS.Auth.html#validate_credentials/1"},{"type":"module","title":"Buckets.Adapters.GCS.AuthServer","doc":"GenServer that manages Google Cloud Storage access tokens.\n\nAutomatically refreshes tokens before they expire to ensure uninterrupted access to GCS APIs.","ref":"Buckets.Adapters.GCS.AuthServer.html"},{"type":"function","title":"Buckets.Adapters.GCS.AuthServer.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Buckets.Adapters.GCS.AuthServer.html#child_spec/1"},{"type":"function","title":"Buckets.Adapters.GCS.AuthServer.get_token/1","doc":"Gets a valid access token from a specific server, refreshing if necessary.","ref":"Buckets.Adapters.GCS.AuthServer.html#get_token/1"},{"type":"function","title":"Buckets.Adapters.GCS.AuthServer.refresh_token/1","doc":"Forces a token refresh.","ref":"Buckets.Adapters.GCS.AuthServer.html#refresh_token/1"},{"type":"function","title":"Buckets.Adapters.GCS.AuthServer.start_link/1","doc":"Starts the token server.\n\nIf opts[:cloud] exists, starts in named mode with the cloud module.\nOtherwise, treats opts as config and starts in nameless mode.","ref":"Buckets.Adapters.GCS.AuthServer.html#start_link/1"},{"type":"module","title":"Buckets.Adapters.GCS.Signature","doc":"Google Cloud Storage V4 signed URL generation.\n\nImplements the V4 signing process for creating signed URLs that allow\ntemporary access to GCS objects without requiring authentication.","ref":"Buckets.Adapters.GCS.Signature.html"},{"type":"function","title":"Buckets.Adapters.GCS.Signature.generate_v4/4","doc":"Generates a V4 signed URL for Google Cloud Storage.","ref":"Buckets.Adapters.GCS.Signature.html#generate_v4/4"},{"type":"function","title":"Options - Buckets.Adapters.GCS.Signature.generate_v4/4","doc":"* `:verb` - HTTP method (default: \"GET\")\n  * `:expires` - Expiration time in seconds from now (default: 3600)\n  * `:headers` - Additional headers to include in the signature","ref":"Buckets.Adapters.GCS.Signature.html#generate_v4/4-options"},{"type":"function","title":"Examples - Buckets.Adapters.GCS.Signature.generate_v4/4","doc":"iex> credentials = %{\"client_email\" => \"test@example.com\", \"private_key\" => \"...\"}\n    iex> Buckets.Adapters.GCS.Signature.generate_v4(credentials, \"my-bucket\", \"path/to/object.jpg\")\n    {:ok, \"https://storage.googleapis.com/my-bucket/path%2Fto%2Fobject.jpg?X-Goog-Algorithm=...\"}","ref":"Buckets.Adapters.GCS.Signature.html#generate_v4/4-examples"},{"type":"module","title":"Buckets.Cloud.Dynamic","doc":"Dynamic configuration management for Cloud modules.\n\nThis module provides functions for managing runtime configuration changes,\nallowing Cloud modules to switch between different adapters and configurations\non a per-process basis, similar to Ecto's dynamic repository functionality.","ref":"Buckets.Cloud.Dynamic.html"},{"type":"module","title":"Usage - Buckets.Cloud.Dynamic","doc":"These functions are automatically available on Cloud modules that use\n`Buckets.Cloud`. You typically don't need to call this module directly.","ref":"Buckets.Cloud.Dynamic.html#module-usage"},{"type":"module","title":"Process-Local Configuration - Buckets.Cloud.Dynamic","doc":"MyApp.Cloud.put_dynamic_config([\n      adapter: Buckets.Adapters.GCS,\n      bucket: \"tenant-bucket\",\n      service_account_credentials: credentials\n    ])","ref":"Buckets.Cloud.Dynamic.html#module-process-local-configuration"},{"type":"module","title":"Scoped Configuration - Buckets.Cloud.Dynamic","doc":"MyApp.Cloud.with_config(config, fn ->\n      MyApp.Cloud.insert(\"file.pdf\")\n    end)","ref":"Buckets.Cloud.Dynamic.html#module-scoped-configuration"},{"type":"module","title":"Auth Server Management - Buckets.Cloud.Dynamic","doc":"For adapters that need background processes (like GCS auth servers), these\nare automatically started and cached per-process as needed through the\nadapter's `prepare_dynamic_config/1` callback.","ref":"Buckets.Cloud.Dynamic.html#module-auth-server-management"},{"type":"function","title":"Buckets.Cloud.Dynamic.config/2","doc":"Gets the current configuration for the given cloud module.\n\nChecks the process dictionary first for dynamic config, falls back to static config.","ref":"Buckets.Cloud.Dynamic.html#config/2"},{"type":"function","title":"Buckets.Cloud.Dynamic.put_dynamic_config/2","doc":"Sets dynamic configuration for the current process.\n\nThis configuration will be used for all subsequent operations in this process\nuntil changed or deleted.","ref":"Buckets.Cloud.Dynamic.html#put_dynamic_config/2"},{"type":"function","title":"Buckets.Cloud.Dynamic.with_config/3","doc":"Executes a function with a specific configuration.\n\nThe configuration is only active during the execution of the function.\nPrevious configuration is restored afterwards.","ref":"Buckets.Cloud.Dynamic.html#with_config/3"},{"type":"module","title":"Buckets.Cloud.Supervisor","doc":"Supervisor that manages any processes required by cloud adapters.\n\nThis supervisor is used internally by Cloud modules to automatically\nstart any required processes for configured adapters by calling the\nadapter's `child_spec/1` callback. Adapters that don't need supervised\nprocesses (like Volume, S3) return `nil`, while adapters that need\nbackground processes (like GCS auth servers) return proper child specs.","ref":"Buckets.Cloud.Supervisor.html"},{"type":"function","title":"Buckets.Cloud.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Buckets.Cloud.Supervisor.html#child_spec/1"},{"type":"function","title":"Buckets.Cloud.Supervisor.start_link/1","doc":"Starts the supervisor with a Cloud module.","ref":"Buckets.Cloud.Supervisor.html#start_link/1"},{"type":"module","title":"Buckets.Location.NotConfigured","doc":"Represents an object that hasn't been assigned a storage location yet.\n\nThis struct is used as the default location for newly created `Buckets.Object`\ninstances that haven't been uploaded to any storage backend. It serves as a\nmarker to indicate that the object exists only in memory or as a local file,\nbut hasn't been persisted to cloud storage.","ref":"Buckets.Location.NotConfigured.html"},{"type":"module","title":"Usage - Buckets.Location.NotConfigured","doc":"Objects with a `NotConfigured` location:\n- Cannot be read from remote storage\n- Cannot generate signed URLs\n- Must be inserted into a Cloud module before remote operations","ref":"Buckets.Location.NotConfigured.html#module-usage"},{"type":"module","title":"Examples - Buckets.Location.NotConfigured","doc":"# New objects start with NotConfigured location\n    object = Buckets.Object.new(\"123\", \"document.pdf\")\n    # object.location == %Buckets.Location.NotConfigured{}\n\n    # After insertion, location is configured\n    {:ok, stored} = MyApp.Cloud.insert(object)\n    # stored.location == %Buckets.Location{path: \"...\", config: ...}","ref":"Buckets.Location.NotConfigured.html#module-examples"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\n## [1.0.0-rc.2] - 2025-06-24","ref":"changelog.html"},{"type":"extras","title":"Added - Changelog","doc":"- Support for multiple cloud storage providers:\n  - Local filesystem (Volume adapter)\n  - Google Cloud Storage (GCS adapter)\n  - Amazon S3 (S3 adapter)\n  - Cloudflare R2 (via S3 adapter with `provider: :cloudflare`)\n  - DigitalOcean Spaces (via S3 adapter with `provider: :digitalocean`)\n  - Tigris (via S3 adapter with `provider: :tigris`)\n- Multi-cloud support with multiple Cloud modules\n- Dynamic cloud configuration for runtime-determined storage\n- Signed URL support for secure uploads and downloads\n- Phoenix LiveView integration for direct-to-cloud uploads\n- Development router for local file uploads/downloads\n- Telemetry integration for monitoring and metrics\n- Comprehensive test suite with live service testing","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Switched from multiple strategies per cloud to one adapter per cloud approach\n- Renamed Strategy modules to Adapter modules\n- Replaced external dependencies (gcs_signed_url, google_api_storage, goth, ex_aws) with req\n- Renamed `cloudflare_r2` provider to `cloudflare` for consistency\n- Cloud modules now act as supervisors for adapters that need background processes","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Router configuration to allow cloud module in location config\n- Automatic setting of correct GCS options for uploads\n- Warning messages for cloud supervisors with no children","ref":"changelog.html#fixed"},{"type":"extras","title":"Internal - Changelog","doc":"- Added `validate_config` callback to adapter behavior\n- Improved Location struct with derived inspect that ignores :config\n- Enhanced documentation and README with core concepts","ref":"changelog.html#internal"},{"type":"extras","title":"Getting Started","doc":"# Getting Started\n\nWelcome to Buckets! This guide will walk you through setting up Buckets in your Phoenix application.","ref":"getting-started.html"},{"type":"extras","title":"Installation - Getting Started","doc":"Add `buckets` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:buckets, \"~> 1.0.0-rc.2\"}\n  ]\nend\n```\n\nThen run:\n\n```bash\nmix deps.get\n```","ref":"getting-started.html#installation"},{"type":"extras","title":"Quick Start - Getting Started","doc":"","ref":"getting-started.html#quick-start"},{"type":"extras","title":"1. Create a Cloud Module - Getting Started","doc":"Create a Cloud module in your application:\n\n```elixir\ndefmodule MyApp.Cloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n```","ref":"getting-started.html#1-create-a-cloud-module"},{"type":"extras","title":"2. Add to Supervision Tree - Getting Started","doc":"Add your Cloud module to your application's supervision tree:\n\n```elixir\ndef start(_type, _args) do\n  children = [\n    # ... other children\n    MyApp.Cloud\n  ]\n\n  Supervisor.start_link(children, opts)\nend\n```","ref":"getting-started.html#2-add-to-supervision-tree"},{"type":"extras","title":"3. Configure Your Adapter - Getting Started","doc":"For development with local filesystem:\n\n```elixir\n# config/dev.exs\nconfig :my_app, MyApp.Cloud,\n  adapter: Buckets.Adapters.Volume,\n  bucket: \"tmp/buckets\",\n  base_url: \"http://localhost:4000\"\n```\n\nFor production with S3:\n\n```elixir\n# config/prod.exs\nconfig :my_app, MyApp.Cloud,\n  adapter: Buckets.Adapters.S3,\n  bucket: \"my-production-bucket\",\n  region: \"us-east-1\",\n  access_key_id: System.fetch_env!(\"AWS_ACCESS_KEY_ID\"),\n  secret_access_key: System.fetch_env!(\"AWS_SECRET_ACCESS_KEY\")\n```","ref":"getting-started.html#3-configure-your-adapter"},{"type":"extras","title":"4. Basic Usage - Getting Started","doc":"Upload a file:\n\n```elixir\n{:ok, object} = MyApp.Cloud.insert(\"/path/to/file.pdf\")\n```\n\nHandle file uploads from a form:\n\n```elixir\ndef create(conn, %{\"upload\" => upload}) do\n  {:ok, object} = MyApp.Cloud.insert(Buckets.Object.from_upload(upload))\n  # Save object metadata to database\n  # ...\nend\n```","ref":"getting-started.html#4-basic-usage"},{"type":"extras","title":"Next Steps - Getting Started","doc":"- Learn about [Core Concepts](core-concepts.html)\n- Explore different [Storage Adapters](../adapters/volume-adapter.html)\n- Set up [Direct Uploads with LiveView](../howtos/direct-uploads-liveview.html)","ref":"getting-started.html#next-steps"},{"type":"extras","title":"Core Concepts","doc":"# Core Concepts\n\nUnderstanding these core concepts will help you use Buckets effectively.","ref":"core-concepts.html"},{"type":"extras","title":"Cloud Modules - Core Concepts","doc":"A Cloud module is your main interface to cloud storage, similar to how an Ecto Repo is your interface to a database. You define Cloud modules in your application:\n\n```elixir\ndefmodule MyApp.Cloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n```\n\nEach Cloud module:\n- Connects to one storage backend at a time\n- Can be configured differently per environment\n- Provides a consistent API regardless of the storage backend","ref":"core-concepts.html#cloud-modules"},{"type":"extras","title":"Adapters - Core Concepts","doc":"Adapters implement the actual storage operations for different backends:\n\n- **`Buckets.Adapters.Volume`** - Local filesystem (great for development)\n- **`Buckets.Adapters.S3`** - Amazon S3 and compatible services\n- **`Buckets.Adapters.GCS`** - Google Cloud Storage\n\nEach adapter has its own configuration requirements and capabilities.","ref":"core-concepts.html#adapters"},{"type":"extras","title":"Objects - Core Concepts","doc":"`Buckets.Object` structs represent files in your system. Objects have:\n\n- **UUID** - Unique identifier\n- **Filename** - Original filename\n- **Data** - The actual file content (when loaded)\n- **Metadata** - Content type, size, and custom metadata\n- **Location** - Where the file is stored\n- **Stored?** - Whether it's been uploaded to cloud storage","ref":"core-concepts.html#objects"},{"type":"extras","title":"Object Lifecycle - Core Concepts","doc":"1. **Created** - Object exists in memory/local disk\n   ```elixir\n   object = Buckets.Object.from_file(\"/tmp/upload.pdf\")\n   # stored?: false, location: NotConfigured\n   ```\n\n2. **Stored** - Object uploaded to cloud storage\n   ```elixir\n   {:ok, stored} = MyApp.Cloud.insert(object)\n   # stored?: true, location: configured\n   ```\n\n3. **Loaded** - Object data fetched from cloud\n   ```elixir\n   {:ok, loaded} = MyApp.Cloud.load(stored)\n   {:ok, data} = Buckets.Object.read(loaded)\n   ```","ref":"core-concepts.html#object-lifecycle"},{"type":"extras","title":"Locations - Core Concepts","doc":"A `Buckets.Location` represents where an object is stored. It contains:\n- **Path** - The storage path (e.g., \"uploads/123/file.pdf\")\n- **Config** - Either adapter configuration or a Cloud module reference","ref":"core-concepts.html#locations"},{"type":"extras","title":"Signed URLs - Core Concepts","doc":"Signed URLs provide temporary, secure access to objects:\n\n```elixir\n{:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 3600)\n# Use signed_url.url for direct access\n```\n\nThese are especially useful for:\n- Direct browser downloads\n- LiveView direct uploads\n- Temporary file sharing","ref":"core-concepts.html#signed-urls"},{"type":"extras","title":"Supervision - Core Concepts","doc":"Some adapters (like GCS) need background processes for authentication. The Cloud module supervisor handles this automatically when added to your supervision tree.\n\nAdapters that need supervision:\n- GCS (for auth token management)\n\nAdapters that don't need supervision:\n- Volume\n- S3","ref":"core-concepts.html#supervision"},{"type":"extras","title":"Configuration - Core Concepts","doc":"Configuration can be:","ref":"core-concepts.html#configuration"},{"type":"extras","title":"Static (compile-time) - Core Concepts","doc":"```elixir\nconfig :my_app, MyApp.Cloud,\n  adapter: Buckets.Adapters.S3,\n  bucket: \"my-bucket\"\n```","ref":"core-concepts.html#static-compile-time"},{"type":"extras","title":"Dynamic (runtime) - Core Concepts","doc":"```elixir\nMyApp.Cloud.put_dynamic_config([\n  adapter: Buckets.Adapters.S3,\n  bucket: \"tenant-bucket\"\n])\n```\n\nThis enables multi-tenant applications where each tenant has their own storage configuration.","ref":"core-concepts.html#dynamic-runtime"},{"type":"extras","title":"Writing Custom Adapters","doc":"# Writing Custom Adapters\n\nThis guide explains how to create your own storage adapter for Buckets.","ref":"writing-custom-adapters.html"},{"type":"extras","title":"Overview - Writing Custom Adapters","doc":"An adapter implements the `Buckets.Adapter` behaviour to provide storage operations for a specific backend. You might create a custom adapter for:\n\n- Proprietary storage systems\n- Specialized cloud providers\n- Hybrid storage solutions\n- Testing purposes","ref":"writing-custom-adapters.html#overview"},{"type":"extras","title":"Basic Structure - Writing Custom Adapters","doc":"```elixir\ndefmodule MyApp.Adapters.CustomStorage do\n  @behaviour Buckets.Adapter\n\n  @impl true\n  def validate_config(config) do\n    # Validate configuration\n  end\n\n  @impl true\n  def put(object, remote_path, config) do\n    # Upload object\n  end\n\n  @impl true\n  def get(remote_path, config) do\n    # Download data\n  end\n\n  @impl true\n  def delete(remote_path, config) do\n    # Delete object\n  end\n\n  @impl true\n  def url(remote_path, config) do\n    # Generate signed URL\n  end\n\n  # Optional - only if you need supervised processes\n  @impl true\n  def child_spec(config, cloud_module) do\n    # Return supervisor child specification\n  end\nend\n```","ref":"writing-custom-adapters.html#basic-structure"},{"type":"extras","title":"Implementing Callbacks - Writing Custom Adapters","doc":"","ref":"writing-custom-adapters.html#implementing-callbacks"},{"type":"extras","title":"validate_config/1 - Writing Custom Adapters","doc":"Validates and normalizes configuration:\n\n```elixir\n@impl true\ndef validate_config(config) do\n  with {:ok, config} <- Keyword.validate(config, [\n         :adapter,\n         :bucket,\n         :api_key,\n         :endpoint,\n         :path\n       ]),\n       {:ok, config} <- Buckets.Adapter.validate_required(config, [\n         :bucket,\n         :api_key,\n         :endpoint\n       ]) do\n    {:ok, config}\n  end\nend\n```","ref":"writing-custom-adapters.html#validate_config-1"},{"type":"extras","title":"put/3 - Writing Custom Adapters","doc":"Uploads an object to storage:\n\n```elixir\n@impl true\ndef put(%Buckets.Object{} = object, remote_path, config) do\n  data = Buckets.Object.read!(object)\n  \n  case upload_to_storage(remote_path, data, config) do\n    :ok -> \n      {:ok, %{}}\n    \n    {:error, reason} -> \n      {:error, reason}\n  end\nend\n```","ref":"writing-custom-adapters.html#put-3"},{"type":"extras","title":"get/2 - Writing Custom Adapters","doc":"Downloads data from storage:\n\n```elixir\n@impl true\ndef get(remote_path, config) do\n  case download_from_storage(remote_path, config) do\n    {:ok, data} -> \n      {:ok, data}\n      \n    {:error, :not_found} -> \n      {:error, :not_found}\n      \n    {:error, reason} -> \n      {:error, reason}\n  end\nend\n```","ref":"writing-custom-adapters.html#get-2"},{"type":"extras","title":"delete/2 - Writing Custom Adapters","doc":"Deletes an object:\n\n```elixir\n@impl true\ndef delete(remote_path, config) do\n  case delete_from_storage(remote_path, config) do\n    :ok -> \n      {:ok, %{}}\n      \n    {:error, reason} -> \n      {:error, reason}\n  end\nend\n```","ref":"writing-custom-adapters.html#delete-2"},{"type":"extras","title":"url/2 - Writing Custom Adapters","doc":"Generates signed URLs:\n\n```elixir\n@impl true\ndef url(remote_path, config) do\n  expires_in = Keyword.get(config, :expires_in, 3600)\n  for_upload = Keyword.get(config, :for_upload, false)\n  \n  signed_url = generate_signed_url(\n    remote_path, \n    expires_in, \n    for_upload,\n    config\n  )\n  \n  location = Buckets.Location.new(remote_path, config)\n  \n  {:ok, %Buckets.SignedURL{\n    url: signed_url,\n    location: location\n  }}\nend\n```","ref":"writing-custom-adapters.html#url-2"},{"type":"extras","title":"child_spec/2 (Optional) - Writing Custom Adapters","doc":"For adapters needing background processes:\n\n```elixir\n@impl true\ndef child_spec(config, cloud_module) do\n  %{\n    id: {__MODULE__, cloud_module},\n    start: {MyApp.AuthServer, :start_link, [[\n      cloud: cloud_module,\n      config: config\n    ]]},\n    restart: :permanent\n  }\nend\n```","ref":"writing-custom-adapters.html#child_spec-2-optional"},{"type":"extras","title":"Best Practices - Writing Custom Adapters","doc":"","ref":"writing-custom-adapters.html#best-practices"},{"type":"extras","title":"1. Configuration Validation - Writing Custom Adapters","doc":"Always validate configuration thoroughly:\n\n```elixir\ndef validate_config(config) do\n  # Use Keyword.validate for known options\n  with {:ok, config} <- Keyword.validate(config, @known_opts),\n       # Check required fields\n       {:ok, config} <- validate_required(config, @required_opts),\n       # Custom validation\n       :ok <- validate_endpoint(config[:endpoint]) do\n    {:ok, config}\n  end\nend\n```","ref":"writing-custom-adapters.html#1-configuration-validation"},{"type":"extras","title":"2. Error Handling - Writing Custom Adapters","doc":"Return consistent error tuples:\n\n```elixir\n# Good\n{:error, :not_found}\n{:error, :unauthorized}\n{:error, {:http_error, 500}}\n\n# Bad\n{:error, \"not found\"}\n:error\nnil\n```","ref":"writing-custom-adapters.html#2-error-handling"},{"type":"extras","title":"3. Metadata Handling - Writing Custom Adapters","doc":"Use object metadata appropriately:\n\n```elixir\ndef put(object, path, config) do\n  headers = build_headers(object.metadata)\n  # Include content-type, cache-control, etc.\nend\n```","ref":"writing-custom-adapters.html#3-metadata-handling"},{"type":"extras","title":"4. Telemetry Integration - Writing Custom Adapters","doc":"Emit telemetry events:\n\n```elixir\ndef put(object, path, config) do\n  metadata = %{\n    adapter: __MODULE__,\n    path: path,\n    size: byte_size(data)\n  }\n  \n  :telemetry.span(\n    [:my_adapter, :put],\n    metadata,\n    fn ->\n      result = do_upload(object, path, config)\n      {result, metadata}\n    end\n  )\nend\n```","ref":"writing-custom-adapters.html#4-telemetry-integration"},{"type":"extras","title":"Testing Your Adapter - Writing Custom Adapters","doc":"Create comprehensive tests:\n\n```elixir\ndefmodule MyApp.Adapters.CustomStorageTest do\n  use ExUnit.Case\n  \n  setup do\n    config = [\n      adapter: MyApp.Adapters.CustomStorage,\n      bucket: \"test-bucket\",\n      api_key: \"test-key\"\n    ]\n    \n    {:ok, config: config}\n  end\n  \n  test \"validates configuration\", %{config: config} do\n    assert {:ok, _} = MyApp.Adapters.CustomStorage.validate_config(config)\n    \n    invalid = Keyword.delete(config, :bucket)\n    assert {:error, [:bucket]} = MyApp.Adapters.CustomStorage.validate_config(invalid)\n  end\n  \n  test \"uploads objects\", %{config: config} do\n    object = Buckets.Object.from_file(\"test/fixtures/sample.pdf\")\n    \n    assert {:ok, _} = MyApp.Adapters.CustomStorage.put(\n      object,\n      \"test/sample.pdf\",\n      config\n    )\n  end\nend\n```","ref":"writing-custom-adapters.html#testing-your-adapter"},{"type":"extras","title":"Example: FTP Adapter - Writing Custom Adapters","doc":"Here's a simplified FTP adapter example:\n\n```elixir\ndefmodule MyApp.Adapters.FTP do\n  @behaviour Buckets.Adapter\n  \n  @impl true\n  def validate_config(config) do\n    with {:ok, config} <- Keyword.validate(config, [\n           :adapter, :host, :port, :username, \n           :password, :bucket, :path\n         ]),\n         {:ok, config} <- validate_required(config, [\n           :host, :username, :password, :bucket\n         ]) do\n      config = Keyword.put_new(config, :port, 21)\n      {:ok, config}\n    end\n  end\n  \n  @impl true\n  def put(object, remote_path, config) do\n    with {:ok, conn} <- connect(config),\n         {:ok, data} <- Buckets.Object.read(object),\n         :ok <- ftp_put(conn, remote_path, data) do\n      {:ok, %{}}\n    end\n  end\n  \n  @impl true\n  def get(remote_path, config) do\n    with {:ok, conn} <- connect(config),\n         {:ok, data} <- ftp_get(conn, remote_path) do\n      {:ok, data}\n    end\n  end\n  \n  # ... implement other callbacks\nend\n```","ref":"writing-custom-adapters.html#example-ftp-adapter"},{"type":"extras","title":"Publishing Your Adapter - Writing Custom Adapters","doc":"1. Create a separate package\n2. Name it `buckets_[provider]` (e.g., `buckets_ftp`)\n3. Add buckets as a dependency\n4. Document configuration options\n5. Provide usage examples\n6. Publish to Hex.pm","ref":"writing-custom-adapters.html#publishing-your-adapter"},{"type":"extras","title":"Need Help? - Writing Custom Adapters","doc":"- Check existing adapters for examples\n- Review the `Buckets.Adapter` behaviour documentation\n- Open an issue for clarification\n- Submit a PR to improve this guide!","ref":"writing-custom-adapters.html#need-help"},{"type":"extras","title":"File Uploads","doc":"# File Uploads\n\nThis guide covers various ways to handle file uploads with Buckets.","ref":"file-uploads.html"},{"type":"extras","title":"Controller Uploads - File Uploads","doc":"The most common scenario is handling file uploads in Phoenix controllers.","ref":"file-uploads.html#controller-uploads"},{"type":"extras","title":"Basic Upload - File Uploads","doc":"```elixir\ndefmodule MyAppWeb.DocumentController do\n  use MyAppWeb, :controller\n  \n  def create(conn, %{\"document\" => %{\"file\" => upload}}) do\n    # Create object from Plug.Upload\n    object = Buckets.Object.from_upload(upload)\n    \n    # Upload to cloud storage\n    case MyApp.Cloud.insert(object) do\n      {:ok, stored_object} ->\n        # Save metadata to database\n        {:ok, document} = Documents.create_document(%{\n          filename: stored_object.filename,\n          path: stored_object.location.path,\n          content_type: stored_object.metadata.content_type,\n          size: stored_object.metadata.content_size\n        })\n        \n        conn\n        |> put_flash(:info, \"Document uploaded successfully\")\n        |> redirect(to: ~p\"/documents/#{document}\")\n        \n      {:error, reason} ->\n        conn\n        |> put_flash(:error, \"Upload failed: #{inspect(reason)}\")\n        |> redirect(to: ~p\"/documents/new\")\n    end\n  end\nend\n```","ref":"file-uploads.html#basic-upload"},{"type":"extras","title":"Multiple File Upload - File Uploads","doc":"```elixir\ndef create(conn, %{\"files\" => uploads}) when is_list(uploads) do\n  results = Enum.map(uploads, fn upload ->\n    object = Buckets.Object.from_upload(upload)\n    MyApp.Cloud.insert(object)\n  end)\n  \n  successful = Enum.filter(results, &match?({:ok, _}, &1))\n  failed = Enum.filter(results, &match?({:error, _}, &1))\n  \n  conn\n  |> put_flash(:info, \"Uploaded #{length(successful)} files\")\n  |> maybe_put_error_flash(failed)\n  |> redirect(to: ~p\"/documents\")\nend\n```","ref":"file-uploads.html#multiple-file-upload"},{"type":"extras","title":"Direct File Upload - File Uploads","doc":"Upload files directly from disk:\n\n```elixir\n# From a file path\n{:ok, object} = MyApp.Cloud.insert(\"/path/to/file.pdf\")\n\n# With custom metadata\nobject = Buckets.Object.new(\n  Ecto.UUID.generate(),\n  \"report.pdf\",\n  metadata: %{\n    author: \"John Doe\",\n    category: \"financial\"\n  }\n)\n\n{:ok, stored} = MyApp.Cloud.insert(object)\n```","ref":"file-uploads.html#direct-file-upload"},{"type":"extras","title":"Post-Upload Processing - File Uploads","doc":"Process files after upload:\n\n```elixir\ndef create(conn, %{\"image\" => upload}) do\n  with object <- Buckets.Object.from_upload(upload),\n       {:ok, stored} <- MyApp.Cloud.insert(object),\n       {:ok, _} <- generate_thumbnails(stored) do\n    # Success\n  else\n    {:error, reason} -> \n      # Handle error\n  end\nend\n\ndefp generate_thumbnails(object) do\n  # Load the image data\n  {:ok, loaded} = MyApp.Cloud.load(object)\n  {:ok, data} = Buckets.Object.read(loaded)\n  \n  # Generate thumbnails\n  sizes = [\n    {:thumb, 150},\n    {:medium, 500},\n    {:large, 1200}\n  ]\n  \n  Enum.each(sizes, fn {name, width} ->\n    resized = resize_image(data, width)\n    \n    thumb_object = Buckets.Object.new(\n      object.uuid,\n      \"#{name}_#{object.filename}\",\n      metadata: %{content_type: \"image/jpeg\"}\n    )\n    \n    MyApp.Cloud.insert(thumb_object)\n  end)\nend\n```","ref":"file-uploads.html#post-upload-processing"},{"type":"extras","title":"Next Steps - File Uploads","doc":"- Learn about [Direct Uploads with LiveView](direct-uploads-liveview.html)\n- Explore [Signed URLs](signed-urls.html) for secure access\n- Set up [Error Handling](error-handling.html) strategies","ref":"file-uploads.html#next-steps"},{"type":"extras","title":"Direct Uploads with LiveView","doc":"# Direct Uploads with LiveView\n\nThis guide shows how to implement direct-to-cloud uploads using Phoenix LiveView, allowing users to upload files directly to your storage provider without going through your server.","ref":"direct-uploads-liveview.html"},{"type":"extras","title":"Overview - Direct Uploads with LiveView","doc":"Direct uploads improve performance and reduce server load by:\n- Uploading files directly from the browser to cloud storage\n- Reducing bandwidth usage on your servers\n- Enabling progress tracking and cancellation\n- Supporting large file uploads","ref":"direct-uploads-liveview.html#overview"},{"type":"extras","title":"Basic Setup - Direct Uploads with LiveView","doc":"","ref":"direct-uploads-liveview.html#basic-setup"},{"type":"extras","title":"1. Configure Your Cloud Module - Direct Uploads with LiveView","doc":"Enable direct uploads by setting the `uploader` option:\n\n```elixir\nconfig :my_app, MyApp.Cloud,\n  adapter: Buckets.Adapters.S3,\n  bucket: \"my-uploads\",\n  uploader: \"S3\",  # or \"GCS\" for Google Cloud Storage\n  # ... other config\n```","ref":"direct-uploads-liveview.html#1-configure-your-cloud-module"},{"type":"extras","title":"2. Basic LiveView Upload - Direct Uploads with LiveView","doc":"```elixir\ndefmodule MyAppWeb.UploadLive do\n  use MyAppWeb, :live_view\n\n  @impl true\n  def mount(_params, _session, socket) do\n    {:ok,\n     socket\n     |> assign(:uploaded_files, [])\n     |> allow_upload(:avatar,\n       accept: ~w(.jpg .jpeg .png),\n       max_entries: 1,\n       max_file_size: 5_000_000,\n       external: &presign_upload/2\n     )}\n  end\n\n  @impl true\n  def render(assigns) do\n    ~H\"\"\"\n     \n      <.live_file_input upload={@uploads.avatar} />\n       Upload \n     \n\n     \n      Uploaded: <%= file.filename %>\n     \n    \"\"\"\n  end\n\n  defp presign_upload(entry, socket) do\n    {:ok, config} = MyApp.Cloud.live_upload(entry)\n    {:ok, config, socket}\n  end\n\n  @impl true\n  def handle_event(\"validate\", _params, socket) do\n    {:noreply, socket}\n  end\n\n  @impl true\n  def handle_event(\"save\", _params, socket) do\n    uploaded_files =\n      consume_uploaded_entries(socket, :avatar, fn %{key: key}, entry ->\n        object = Buckets.Object.from_upload({entry, %{key: key}})\n        {:ok, object}\n      end)\n\n    {:noreply, update(socket, :uploaded_files, &(&1 ++ uploaded_files))}\n  end\nend\n```","ref":"direct-uploads-liveview.html#2-basic-liveview-upload"},{"type":"extras","title":"Multiple File Uploads - Direct Uploads with LiveView","doc":"```elixir\ndef mount(_params, _session, socket) do\n  {:ok,\n   socket\n   |> assign(:uploaded_files, [])\n   |> allow_upload(:documents,\n     accept: ~w(.pdf .doc .docx),\n     max_entries: 10,\n     max_file_size: 10_000_000,\n     external: &presign_upload/2,\n     auto_upload: true\n   )}\nend\n```","ref":"direct-uploads-liveview.html#multiple-file-uploads"},{"type":"extras","title":"Error Handling - Direct Uploads with LiveView","doc":"```elixir\ndef render(assigns) do\n  ~H\"\"\"\n   \n    <%= error_to_string(err) %>\n   \n  \"\"\"\nend\n\ndefp error_to_string(:too_large), do: \"File is too large\"\ndefp error_to_string(:not_accepted), do: \"File type not accepted\"\ndefp error_to_string(:too_many_files), do: \"Too many files\"\ndefp error_to_string(:external_client_failure), do: \"Upload failed\"\n\ndefp presign_upload(entry, socket) do\n  case MyApp.Cloud.live_upload(entry) do\n    {:ok, config} ->\n      {:ok, config, socket}\n      \n    {:error, reason} ->\n      {:error, \"Failed to generate upload URL: #{inspect(reason)}\", socket}\n  end\nend\n```","ref":"direct-uploads-liveview.html#error-handling"},{"type":"extras","title":"CORS Configuration - Direct Uploads with LiveView","doc":"For direct uploads to work, configure CORS on your bucket:","ref":"direct-uploads-liveview.html#cors-configuration"},{"type":"extras","title":"S3 CORS - Direct Uploads with LiveView","doc":"```xml\n<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n \n   \n     https://myapp.com \n     PUT \n     POST \n     * \n     ETag \n     3000 \n   \n \n```","ref":"direct-uploads-liveview.html#s3-cors"},{"type":"extras","title":"GCS CORS - Direct Uploads with LiveView","doc":"```json\n[\n  {\n    \"origin\": [\"https://myapp.com\"],\n    \"method\": [\"PUT\", \"POST\"],\n    \"responseHeader\": [\"*\"],\n    \"maxAgeSeconds\": 3600\n  }\n]\n```","ref":"direct-uploads-liveview.html#gcs-cors"},{"type":"extras","title":"Signed URLs","doc":"# Signed URLs\n\nSigned URLs provide temporary, secure access to objects in cloud storage without requiring authentication. This guide explains how to generate and use signed URLs with Buckets.","ref":"signed-urls.html"},{"type":"extras","title":"What are Signed URLs? - Signed URLs","doc":"Signed URLs are time-limited URLs that grant temporary access to private objects. They include:\n- The object location\n- Expiration time\n- Cryptographic signature\n- Optional permissions (read/write)","ref":"signed-urls.html#what-are-signed-urls"},{"type":"extras","title":"Basic Usage - Signed URLs","doc":"","ref":"signed-urls.html#basic-usage"},{"type":"extras","title":"Generating Download URLs - Signed URLs","doc":"```elixir\n# Generate a signed URL for downloads (default: 1 hour)\n{:ok, signed_url} = MyApp.Cloud.url(object)\n\n# Access the URL string\ndownload_url = signed_url.url\n\n# Or use String protocol\ndownload_url = to_string(signed_url)\n```","ref":"signed-urls.html#generating-download-urls"},{"type":"extras","title":"Custom Expiration - Signed URLs","doc":"```elixir\n# Expires in 5 minutes\n{:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 300)\n\n# Expires in 24 hours\n{:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 86400)\n\n# Expires in 7 days (maximum for S3)\n{:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 604800)\n```","ref":"signed-urls.html#custom-expiration"},{"type":"extras","title":"Upload URLs - Signed URLs","doc":"Generate URLs for direct uploads:\n\n```elixir\n# Create a placeholder object\nobject = Buckets.Object.new(\n  Ecto.UUID.generate(),\n  \"document.pdf\",\n  location: {\"uploads/documents/document.pdf\", MyApp.Cloud}\n)\n\n# Generate upload URL\n{:ok, upload_url} = MyApp.Cloud.url(object, \n  expires_in: 3600,\n  for_upload: true\n)\n```","ref":"signed-urls.html#upload-urls"},{"type":"extras","title":"Use Cases - Signed URLs","doc":"","ref":"signed-urls.html#use-cases"},{"type":"extras","title":"Secure File Downloads - Signed URLs","doc":"```elixir\ndefmodule MyAppWeb.DocumentController do\n  def download(conn, %{\"id\" => id}) do\n    document = Documents.get!(id)\n    \n    # Load object from storage location\n    object = Buckets.Object.new(\n      document.id,\n      document.filename,\n      location: {document.storage_path, MyApp.Cloud}\n    )\n    \n    # Generate temporary download URL\n    {:ok, signed_url} = MyApp.Cloud.url(object, expires_in: 300)\n    \n    # Redirect to signed URL\n    redirect(conn, external: signed_url.url)\n  end\nend\n```","ref":"signed-urls.html#secure-file-downloads"},{"type":"extras","title":"Email Attachments - Signed URLs","doc":"```elixir\ndefmodule MyApp.Mailer do\n  def invoice_email(user, invoice) do\n    # Generate URL valid for 7 days\n    {:ok, download_url} = MyApp.Cloud.url(invoice.document, \n      expires_in: 604800\n    )\n    \n    new()\n    |> to(user.email)\n    |> subject(\"Your Invoice\")\n    |> html_body(\"\"\"\n     Your invoice is ready! \n      Download Invoice  \n     This link expires in 7 days. \n    \"\"\")\n    |> deliver()\n  end\nend\n```","ref":"signed-urls.html#email-attachments"},{"type":"extras","title":"API Responses - Signed URLs","doc":"```elixir\ndefmodule MyAppWeb.API.FileController do\n  def show(conn, %{\"id\" => id}) do\n    file = Files.get!(id)\n    \n    # Generate short-lived URL\n    {:ok, signed_url} = MyApp.Cloud.url(file.object, \n      expires_in: 600  # 10 minutes\n    )\n    \n    json(conn, %{\n      id: file.id,\n      filename: file.filename,\n      size: file.size,\n      content_type: file.content_type,\n      download_url: signed_url.url,\n      expires_at: DateTime.add(DateTime.utc_now(), 600)\n    })\n  end\nend\n```","ref":"signed-urls.html#api-responses"},{"type":"extras","title":"Multi-Cloud Setup","doc":"# Multi-Cloud Setup\n\nThis guide explains how to use multiple cloud storage providers simultaneously in your application.","ref":"multi-cloud-setup.html"},{"type":"extras","title":"Why Multi-Cloud? - Multi-Cloud Setup","doc":"Multi-cloud setups are useful for:\n- **Geographic distribution** - Store files closer to users\n- **Cost optimization** - Use the most cost-effective provider for each use case\n- **Redundancy** - Failover between providers\n- **Compliance** - Meet data residency requirements\n- **Migration** - Gradually move between providers","ref":"multi-cloud-setup.html#why-multi-cloud"},{"type":"extras","title":"Basic Multi-Cloud Setup - Multi-Cloud Setup","doc":"","ref":"multi-cloud-setup.html#basic-multi-cloud-setup"},{"type":"extras","title":"1. Define Multiple Cloud Modules - Multi-Cloud Setup","doc":"```elixir\n# For local development\ndefmodule MyApp.LocalCloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n\n# For user uploads\ndefmodule MyApp.S3Cloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n\n# For backups and archives\ndefmodule MyApp.GCSCloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n\n# For CDN assets\ndefmodule MyApp.R2Cloud do\n  use Buckets.Cloud, otp_app: :my_app\nend\n```","ref":"multi-cloud-setup.html#1-define-multiple-cloud-modules"},{"type":"extras","title":"2. Configure Each Cloud - Multi-Cloud Setup","doc":"```elixir\n# config/dev.exs\nconfig :my_app, MyApp.LocalCloud,\n  adapter: Buckets.Adapters.Volume,\n  bucket: \"tmp/local_storage\",\n  base_url: \"http://localhost:4000\"\n\n# config/config.exs or runtime.exs\nconfig :my_app, MyApp.S3Cloud,\n  adapter: Buckets.Adapters.S3,\n  bucket: System.get_env(\"S3_BUCKET\"),\n  region: \"us-east-1\",\n  access_key_id: System.get_env(\"AWS_ACCESS_KEY_ID\"),\n  secret_access_key: System.get_env(\"AWS_SECRET_ACCESS_KEY\")\n\nconfig :my_app, MyApp.GCSCloud,\n  adapter: Buckets.Adapters.GCS,\n  bucket: System.get_env(\"GCS_BUCKET\"),\n  service_account_credentials: System.get_env(\"GOOGLE_CREDENTIALS\")\n\nconfig :my_app, MyApp.R2Cloud,\n  adapter: Buckets.Adapters.S3,\n  provider: :cloudflare,\n  bucket: System.get_env(\"R2_BUCKET\"),\n  endpoint_url: System.get_env(\"R2_ENDPOINT\"),\n  access_key_id: System.get_env(\"R2_ACCESS_KEY\"),\n  secret_access_key: System.get_env(\"R2_SECRET_KEY\")\n```","ref":"multi-cloud-setup.html#2-configure-each-cloud"},{"type":"extras","title":"3. Add to Supervision Tree - Multi-Cloud Setup","doc":"```elixir\ndef start(_type, _args) do\n  children = [\n    # ... other processes\n    MyApp.GCSCloud,  # Only GCS needs supervision\n    # S3, R2, and Volume don't need supervision\n  ]\n  \n  Supervisor.start_link(children, opts)\nend\n```","ref":"multi-cloud-setup.html#3-add-to-supervision-tree"},{"type":"extras","title":"Usage Patterns - Multi-Cloud Setup","doc":"","ref":"multi-cloud-setup.html#usage-patterns"},{"type":"extras","title":"Content-Based Routing - Multi-Cloud Setup","doc":"Route files to different clouds based on content:\n\n```elixir\ndefmodule MyApp.Storage do\n  def store_file(upload) do\n    cloud = determine_cloud(upload)\n    object = Buckets.Object.from_upload(upload)\n    cloud.insert(object)\n  end\n  \n  defp determine_cloud(upload) do\n    case upload.content_type do\n      \"image/\" <> _ -> MyApp.R2Cloud      # Images to CDN\n      \"video/\" <> _ -> MyApp.S3Cloud      # Videos to S3\n      \"application/pdf\" -> MyApp.GCSCloud # Documents to GCS\n      _ -> MyApp.S3Cloud                  # Default to S3\n    end\n  end\nend\n```","ref":"multi-cloud-setup.html#content-based-routing"},{"type":"extras","title":"Geographic Distribution - Multi-Cloud Setup","doc":"Store files in the closest region:\n\n```elixir\ndefmodule MyApp.GeoStorage do\n  @clouds %{\n    \"US\" => MyApp.USCloud,\n    \"EU\" => MyApp.EUCloud,\n    \"ASIA\" => MyApp.AsiaCloud\n  }\n  \n  def store_for_user(upload, user) do\n    cloud = @clouds[user.region] || MyApp.USCloud\n    object = Buckets.Object.from_upload(upload)\n    cloud.insert(object)\n  end\nend\n```","ref":"multi-cloud-setup.html#geographic-distribution"},{"type":"extras","title":"Cross-Cloud Operations - Multi-Cloud Setup","doc":"","ref":"multi-cloud-setup.html#cross-cloud-operations"},{"type":"extras","title":"Copying Between Clouds - Multi-Cloud Setup","doc":"```elixir\ndefmodule MyApp.CloudSync do\n  def copy_object(object, from_cloud, to_cloud) do\n    # Load from source\n    {:ok, loaded} = from_cloud.load(object)\n    \n    # Insert to destination\n    {:ok, copied} = to_cloud.insert(loaded)\n    \n    copied\n  end\n  \n  def migrate_bucket(from_cloud, to_cloud) do\n    objects = list_all_objects(from_cloud)\n    \n    objects\n    |> Task.async_stream(\n      fn object ->\n        copy_object(object, from_cloud, to_cloud)\n      end,\n      max_concurrency: 10,\n      timeout: 60_000\n    )\n    |> Stream.run()\n  end\nend\n```","ref":"multi-cloud-setup.html#copying-between-clouds"},{"type":"extras","title":"Database Design - Multi-Cloud Setup","doc":"Store cloud information with your files:\n\n```elixir\ndefmodule MyApp.Files.File do\n  use Ecto.Schema\n  \n  schema \"files\" do\n    field :filename, :string\n    field :cloud_module, :string  # \"Elixir.MyApp.S3Cloud\"\n    field :storage_path, :string\n    field :content_type, :string\n    field :size, :integer\n    field :region, :string\n    \n    timestamps()\n  end\n  \n  def to_object(%__MODULE__{} = file) do\n    cloud = String.to_existing_atom(file.cloud_module)\n    \n    Buckets.Object.new(\n      file.id,\n      file.filename,\n      location: {file.storage_path, cloud},\n      metadata: %{\n        content_type: file.content_type,\n        content_size: file.size\n      }\n    )\n  end\nend\n```","ref":"multi-cloud-setup.html#database-design"},{"type":"extras","title":"Dynamic Configuration","doc":"# Dynamic Configuration\n\nThis guide explains how to configure cloud storage dynamically at runtime, enabling multi-tenant applications and flexible storage strategies.","ref":"dynamic-configuration.html"},{"type":"extras","title":"Overview - Dynamic Configuration","doc":"Dynamic configuration allows you to:\n- Support multi-tenant applications with per-tenant storage\n- Switch storage providers without restarting\n- Test with different configurations\n- Implement feature flags for storage","ref":"dynamic-configuration.html#overview"},{"type":"extras","title":"Basic Dynamic Configuration - Dynamic Configuration","doc":"","ref":"dynamic-configuration.html#basic-dynamic-configuration"},{"type":"extras","title":"Process-Scoped Configuration - Dynamic Configuration","doc":"Set configuration for the current process:\n\n```elixir\n# Set dynamic config\nMyApp.Cloud.put_dynamic_config([\n  adapter: Buckets.Adapters.S3,\n  bucket: \"tenant-123-bucket\",\n  region: \"us-west-2\",\n  access_key_id: \"AKIA...\",\n  secret_access_key: \"secret...\"\n])\n\n# All subsequent operations use this config\n{:ok, object} = MyApp.Cloud.insert(upload)\n```","ref":"dynamic-configuration.html#process-scoped-configuration"},{"type":"extras","title":"Scoped Configuration Blocks - Dynamic Configuration","doc":"Use configuration for a specific block of code:\n\n```elixir\nconfig = [\n  adapter: Buckets.Adapters.GCS,\n  bucket: \"temporary-bucket\",\n  service_account_credentials: credentials\n]\n\nresult = MyApp.Cloud.with_config(config, fn ->\n  # All operations in this block use the config\n  {:ok, obj1} = MyApp.Cloud.insert(file1)\n  {:ok, obj2} = MyApp.Cloud.insert(file2)\n  \n  [obj1, obj2]\nend)\n```","ref":"dynamic-configuration.html#scoped-configuration-blocks"},{"type":"extras","title":"Multi-Tenant Implementation - Dynamic Configuration","doc":"","ref":"dynamic-configuration.html#multi-tenant-implementation"},{"type":"extras","title":"Basic Multi-Tenant Setup - Dynamic Configuration","doc":"```elixir\ndefmodule MyApp.TenantStorage do\n  def with_tenant(tenant, fun) do\n    config = build_config(tenant)\n    MyApp.Cloud.with_config(config, fun)\n  end\n  \n  defp build_config(tenant) do\n    case tenant.storage_provider do\n      \"s3\" ->\n        [\n          adapter: Buckets.Adapters.S3,\n          bucket: tenant.s3_bucket,\n          region: tenant.s3_region,\n          access_key_id: decrypt(tenant.s3_access_key),\n          secret_access_key: decrypt(tenant.s3_secret_key)\n        ]\n        \n      \"gcs\" ->\n        [\n          adapter: Buckets.Adapters.GCS,\n          bucket: tenant.gcs_bucket,\n          service_account_credentials: decrypt(tenant.gcs_credentials)\n        ]\n        \n      \"volume\" ->\n        [\n          adapter: Buckets.Adapters.Volume,\n          bucket: \"tenants/#{tenant.id}\",\n          base_url: tenant.base_url || \"http://localhost:4000\"\n        ]\n    end\n  end\n  \n  defp decrypt(encrypted_value) do\n    # Decrypt sensitive credentials\n    MyApp.Crypto.decrypt(encrypted_value)\n  end\nend\n```","ref":"dynamic-configuration.html#basic-multi-tenant-setup"},{"type":"extras","title":"Phoenix Integration - Dynamic Configuration","doc":"Use dynamic config in Phoenix controllers:\n\n```elixir\ndefmodule MyAppWeb.FileController do\n  plug :load_tenant_config\n  \n  def upload(conn, %{\"file\" => upload}) do\n    # Config is already set by plug\n    object = Buckets.Object.from_upload(upload)\n    \n    case MyApp.Cloud.insert(object) do\n      {:ok, stored} ->\n        conn\n        |> put_flash(:info, \"File uploaded\")\n        |> redirect(to: ~p\"/files\")\n        \n      {:error, reason} ->\n        conn\n        |> put_flash(:error, \"Upload failed: #{inspect(reason)}\")\n        |> redirect(to: ~p\"/files/new\")\n    end\n  end\n  \n  defp load_tenant_config(conn, _opts) do\n    tenant = conn.assigns.current_tenant\n    config = MyApp.TenantStorage.build_config(tenant)\n    \n    MyApp.Cloud.put_dynamic_config(config)\n    conn\n  end\nend\n```","ref":"dynamic-configuration.html#phoenix-integration"},{"type":"extras","title":"LiveView Integration - Dynamic Configuration","doc":"```elixir\ndefmodule MyAppWeb.UploadLive do\n  use MyAppWeb, :live_view\n  \n  def mount(_params, session, socket) do\n    tenant = get_tenant(session)\n    config = MyApp.TenantStorage.build_config(tenant)\n    \n    # Set config for this LiveView process\n    MyApp.Cloud.put_dynamic_config(config)\n    \n    {:ok,\n     socket\n     |> assign(:tenant, tenant)\n     |> allow_upload(:files, accept: :any, external: &presign_upload/2)}\n  end\n  \n  defp presign_upload(entry, socket) do\n    # Uses the dynamic config set in mount\n    {:ok, config} = MyApp.Cloud.live_upload(entry)\n    {:ok, config, socket}\n  end\nend\n```","ref":"dynamic-configuration.html#liveview-integration"},{"type":"extras","title":"Background Jobs - Dynamic Configuration","doc":"","ref":"dynamic-configuration.html#background-jobs"},{"type":"extras","title":"Oban Integration - Dynamic Configuration","doc":"```elixir\ndefmodule MyApp.FileProcessor do\n  use Oban.Worker\n  \n  @impl true\n  def perform(%{args: %{\"file_id\" => file_id, \"tenant_id\" => tenant_id}}) do\n    tenant = Tenants.get!(tenant_id)\n    file = Files.get!(file_id)\n    \n    # Run with tenant's config\n    MyApp.TenantStorage.with_tenant(tenant, fn ->\n      # Load file data\n      {:ok, object} = MyApp.Cloud.load(file.object)\n      \n      # Process file\n      process_file(object)\n      \n      # Save results\n      {:ok, processed} = MyApp.Cloud.insert(processed_object)\n      \n      Files.update(file, %{processed_path: processed.location.path})\n    end)\n    \n    :ok\n  end\nend\n```","ref":"dynamic-configuration.html#oban-integration"},{"type":"extras","title":"Error Handling","doc":"# Error Handling\n\nThis guide covers error handling strategies for robust file storage operations with Buckets.","ref":"error-handling.html"},{"type":"extras","title":"Error Types - Error Handling","doc":"","ref":"error-handling.html#error-types"},{"type":"extras","title":"Common Errors - Error Handling","doc":"Buckets operations can return various error types:\n\n```elixir\n# Configuration errors\n{:error, [:bucket, :access_key_id]}  # Missing required fields\n\n# Network errors\n{:error, %Mint.TransportError{}}     # Connection failed\n{:error, {:http_error, 500}}         # Server error\n\n# Permission errors\n{:error, :unauthorized}              # Invalid credentials\n{:error, :forbidden}                 # Insufficient permissions\n\n# Resource errors\n{:error, :not_found}                 # Object doesn't exist\n{:error, :already_exists}            # Duplicate object\n\n# Validation errors\n{:error, :file_too_large}            # Size limit exceeded\n{:error, :invalid_content_type}      # Type not allowed\n```","ref":"error-handling.html#common-errors"},{"type":"extras","title":"Basic Error Handling - Error Handling","doc":"","ref":"error-handling.html#basic-error-handling"},{"type":"extras","title":"Pattern Matching - Error Handling","doc":"```elixir\ncase MyApp.Cloud.insert(object) do\n  {:ok, stored} ->\n    # Success path\n    {:ok, stored}\n    \n  {:error, :unauthorized} ->\n    # Handle auth error\n    Logger.error(\"Invalid cloud storage credentials\")\n    {:error, \"Authentication failed\"}\n    \n  {:error, %{status: 507}} ->\n    # Storage full\n    Logger.error(\"Cloud storage quota exceeded\")\n    {:error, \"Storage limit reached\"}\n    \n  {:error, reason} ->\n    # Generic error\n    Logger.error(\"Upload failed: #{inspect(reason)}\")\n    {:error, \"Upload failed\"}\nend\n```","ref":"error-handling.html#pattern-matching"},{"type":"extras","title":"Using Bang Functions - Error Handling","doc":"```elixir\ndef upload_or_raise(file_path) do\n  try do\n    object = Buckets.Object.from_file(file_path)\n    stored = MyApp.Cloud.insert!(object)\n    {:ok, stored}\n  rescue\n    e in RuntimeError ->\n      Logger.error(\"Upload failed: #{e.message}\")\n      reraise e, __STACKTRACE__\n  end\nend\n```","ref":"error-handling.html#using-bang-functions"},{"type":"extras","title":"Retry Strategies - Error Handling","doc":"","ref":"error-handling.html#retry-strategies"},{"type":"extras","title":"Simple Retry - Error Handling","doc":"```elixir\ndefmodule MyApp.Storage.Retry do\n  def with_retry(fun, opts \\\\ []) do\n    max_attempts = Keyword.get(opts, :max_attempts, 3)\n    delay = Keyword.get(opts, :delay, 1000)\n    \n    do_retry(fun, max_attempts, delay, 1)\n  end\n  \n  defp do_retry(fun, max_attempts, delay, attempt) do\n    case fun.() do\n      {:ok, result} ->\n        {:ok, result}\n        \n      {:error, reason} when attempt  \n        if retryable_error?(reason) do\n          Process.sleep(delay * attempt)  # Exponential backoff\n          do_retry(fun, max_attempts, delay, attempt + 1)\n        else\n          {:error, reason}\n        end\n        \n      {:error, reason} ->\n        {:error, reason}\n    end\n  end\n  \n  defp retryable_error?(:timeout), do: true\n  defp retryable_error?({:http_error, status}) when status >= 500, do: true\n  defp retryable_error?(%Mint.TransportError{}), do: true\n  defp retryable_error?(_), do: false\nend\n\n# Usage\nMyApp.Storage.Retry.with_retry(fn ->\n  MyApp.Cloud.insert(object)\nend, max_attempts: 5, delay: 2000)\n```","ref":"error-handling.html#simple-retry"},{"type":"extras","title":"User-Friendly Errors - Error Handling","doc":"","ref":"error-handling.html#user-friendly-errors"},{"type":"extras","title":"Error Translation - Error Handling","doc":"```elixir\ndefmodule MyApp.Storage.ErrorMessages do\n  def translate({:error, :unauthorized}) do\n    \"Storage authentication failed. Please check your credentials.\"\n  end\n  \n  def translate({:error, :not_found}) do\n    \"The requested file could not be found.\"\n  end\n  \n  def translate({:error, {:http_error, 507}}) do\n    \"Storage quota exceeded. Please upgrade your plan.\"\n  end\n  \n  def translate({:error, :timeout}) do\n    \"The operation timed out. Please try again.\"\n  end\n  \n  def translate({:error, :file_too_large}) do\n    \"The file is too large. Maximum size is 10MB.\"\n  end\n  \n  def translate(_error) do\n    \"An unexpected error occurred. Please try again later.\"\n  end\nend\n\n# In controller\ncase MyApp.Cloud.insert(object) do\n  {:ok, _} ->\n    put_flash(conn, :info, \"Upload successful\")\n    \n  {:error, _} = error ->\n    message = MyApp.Storage.ErrorMessages.translate(error)\n    put_flash(conn, :error, message)\nend\n```","ref":"error-handling.html#error-translation"},{"type":"extras","title":"Testing","doc":"# Testing\n\nThis guide covers testing strategies for applications using Buckets, including unit tests, integration tests, and testing with live cloud services.","ref":"testing.html"},{"type":"extras","title":"Test Setup - Testing","doc":"","ref":"testing.html#test-setup"},{"type":"extras","title":"Basic Test Configuration - Testing","doc":"```elixir\n# config/test.exs\nconfig :my_app, MyApp.Cloud,\n  adapter: Buckets.Adapters.Volume,\n  bucket: \"tmp/test_storage\",\n  base_url: \"http://localhost:4001\"\n\n# Ensure test isolation\nconfig :my_app, :storage_base_path, \"tmp/test_#{System.get_pid()}\"\n```","ref":"testing.html#basic-test-configuration"},{"type":"extras","title":"Test Helper Module - Testing","doc":"```elixir\ndefmodule MyApp.StorageCase do\n  use ExUnit.CaseTemplate\n  \n  using do\n    quote do\n      import MyApp.StorageCase\n      \n      setup :create_test_storage\n    end\n  end\n  \n  def create_test_storage(_context) do\n    # Create unique test directory\n    test_dir = \"tmp/test_#{:rand.uniform(1_000_000)}\"\n    File.mkdir_p!(test_dir)\n    \n    # Configure for this test\n    config = [\n      adapter: Buckets.Adapters.Volume,\n      bucket: test_dir,\n      base_url: \"http://localhost:4001\"\n    ]\n    \n    MyApp.Cloud.put_dynamic_config(config)\n    \n    on_exit(fn ->\n      # Cleanup\n      File.rm_rf!(test_dir)\n    end)\n    \n    {:ok, storage_dir: test_dir}\n  end\n  \n  def create_test_object(attrs \\\\ %{}) do\n    defaults = %{\n      uuid: Ecto.UUID.generate(),\n      filename: \"test.pdf\",\n      content: \"test content\",\n      content_type: \"application/pdf\"\n    }\n    \n    attrs = Map.merge(defaults, attrs)\n    \n    Buckets.Object.new(\n      attrs.uuid,\n      attrs.filename,\n      metadata: %{\n        content_type: attrs.content_type,\n        content_size: byte_size(attrs.content)\n      }\n    )\n  end\nend\n```","ref":"testing.html#test-helper-module"},{"type":"extras","title":"Unit Tests - Testing","doc":"","ref":"testing.html#unit-tests"},{"type":"extras","title":"Testing Storage Operations - Testing","doc":"```elixir\ndefmodule MyApp.CloudTest do\n  use MyApp.StorageCase\n  \n  describe \"insert/1\" do\n    test \"stores object successfully\", %{storage_dir: dir} do\n      object = create_test_object()\n      \n      assert {:ok, stored} = MyApp.Cloud.insert(object)\n      assert stored.stored?\n      assert stored.location.path =~ object.uuid\n      \n      # Verify file exists\n      file_path = Path.join([dir, stored.location.path])\n      assert File.exists?(file_path)\n    end\n    \n    test \"handles missing file\" do\n      object = Buckets.Object.from_file(\"non_existent.pdf\")\n      \n      assert_raise RuntimeError, ~r/non-existent file/, fn ->\n        MyApp.Cloud.insert!(object)\n      end\n    end\n  end\n  \n  describe \"read/1\" do\n    test \"reads stored object data\" do\n      object = create_test_object(content: \"Hello, World!\")\n      {:ok, stored} = MyApp.Cloud.insert(object)\n      \n      assert {:ok, data} = MyApp.Cloud.read(stored)\n      assert data == \"Hello, World!\"\n    end\n    \n    test \"returns error for missing object\" do\n      fake_object = create_test_object()\n      |> Map.put(:location, Buckets.Location.new(\"fake/path\", MyApp.Cloud))\n      |> Map.put(:stored?, true)\n      \n      assert {:error, :not_found} = MyApp.Cloud.read(fake_object)\n    end\n  end\nend\n```","ref":"testing.html#testing-storage-operations"},{"type":"extras","title":"LiveView Testing - Testing","doc":"","ref":"testing.html#liveview-testing"},{"type":"extras","title":"Testing Direct Uploads - Testing","doc":"```elixir\ndefmodule MyAppWeb.UploadLiveTest do\n  use MyAppWeb.ConnCase\n  use MyApp.StorageCase\n  import Phoenix.LiveViewTest\n  \n  describe \"file uploads\" do\n    test \"direct upload to cloud storage\", %{conn: conn} do\n      {:ok, view, _html} = live(conn, ~p\"/uploads/new\")\n      \n      # Simulate file selection\n      file =\n        file_input(view, \"form\", :document, [\n          %{\n            name: \"invoice.pdf\",\n            content: File.read!(\"test/fixtures/sample.pdf\"),\n            type: \"application/pdf\"\n          }\n        ])\n      \n      # Trigger upload\n      assert render_upload(file, \"invoice.pdf\")\n      \n      # Submit form\n      html = form(view, \"form\") |> render_submit()\n      \n      assert html =~ \"Upload successful\"\n      assert MyApp.Files.count() == 1\n    end\n  end\nend\n```","ref":"testing.html#testing-direct-uploads"},{"type":"extras","title":"Mocking and Stubbing - Testing","doc":"","ref":"testing.html#mocking-and-stubbing"},{"type":"extras","title":"Using Mox for Cloud Operations - Testing","doc":"```elixir\n# test/support/mocks.ex\nMox.defmock(MyApp.CloudMock, for: Buckets.Cloud)\n\n# In tests\ndefmodule MyApp.ServiceTest do\n  use ExUnit.Case\n  import Mox\n  \n  setup :verify_on_exit!\n  \n  test \"processes file after upload\" do\n    object = create_test_object()\n    \n    expect(MyApp.CloudMock, :insert, fn ^object ->\n      {:ok, %{object | stored?: true}}\n    end)\n    \n    expect(MyApp.CloudMock, :read, fn _object ->\n      {:ok, \"processed content\"}\n    end)\n    \n    assert {:ok, result} = MyApp.FileProcessor.process(object)\n    assert result.content == \"processed content\"\n  end\nend\n```","ref":"testing.html#using-mox-for-cloud-operations"},{"type":"extras","title":"Running Tests - Testing","doc":"```bash\n# Run all tests\nmix test\n\n# Run only unit tests (exclude external)\nmix test --exclude external\n\n# Run specific adapter tests\nmix test --only s3\n\n# Run with coverage\nmix test --cover\n```","ref":"testing.html#running-tests"},{"type":"extras","title":"Telemetry and Monitoring","doc":"# Telemetry and Monitoring\n\nThis guide covers monitoring and observability for Buckets operations using Telemetry.","ref":"telemetry-and-monitoring.html"},{"type":"extras","title":"Overview - Telemetry and Monitoring","doc":"Buckets emits telemetry events for all major operations, allowing you to monitor performance, track errors, and gather metrics about your file storage operations.","ref":"telemetry-and-monitoring.html#overview"},{"type":"extras","title":"Available Events - Telemetry and Monitoring","doc":"","ref":"telemetry-and-monitoring.html#available-events"},{"type":"extras","title":"Cloud Operations - Telemetry and Monitoring","doc":"- `[:buckets, :cloud, :insert, :start]` - File upload started\n- `[:buckets, :cloud, :insert, :stop]` - File upload completed\n- `[:buckets, :cloud, :insert, :exception]` - File upload failed\n- `[:buckets, :cloud, :read, :start]` - File read started\n- `[:buckets, :cloud, :read, :stop]` - File read completed\n- `[:buckets, :cloud, :delete, :start]` - File deletion started\n- `[:buckets, :cloud, :delete, :stop]` - File deletion completed\n- `[:buckets, :cloud, :url, :start]` - URL generation started\n- `[:buckets, :cloud, :url, :stop]` - URL generation completed","ref":"telemetry-and-monitoring.html#cloud-operations"},{"type":"extras","title":"Adapter Operations - Telemetry and Monitoring","doc":"- `[:buckets, :adapter, :put, :start]` - Adapter upload started\n- `[:buckets, :adapter, :put, :stop]` - Adapter upload completed\n- `[:buckets, :adapter, :get, :start]` - Adapter download started\n- `[:buckets, :adapter, :get, :stop]` - Adapter download completed","ref":"telemetry-and-monitoring.html#adapter-operations"},{"type":"extras","title":"Basic Setup - Telemetry and Monitoring","doc":"```elixir\ndefmodule MyApp.Telemetry do\n  require Logger\n\n  def setup do\n    events = [\n      [:buckets, :cloud, :insert, :start],\n      [:buckets, :cloud, :insert, :stop],\n      [:buckets, :cloud, :insert, :exception],\n      [:buckets, :cloud, :read, :stop],\n      [:buckets, :cloud, :delete, :stop],\n      [:buckets, :cloud, :url, :stop]\n    ]\n\n    :telemetry.attach_many(\n      \"buckets-handler\",\n      events,\n      &handle_event/4,\n      nil\n    )\n  end\n\n  defp handle_event([:buckets, :cloud, :insert, :stop], measurements, metadata, _config) do\n    Logger.info(\"File uploaded\",\n      duration_ms: measurements.duration / 1_000_000,\n      filename: metadata.filename,\n      cloud: metadata.cloud_module\n    )\n  end\n\n  defp handle_event([:buckets, :cloud, :insert, :exception], _measurements, metadata, _config) do\n    Logger.error(\"File upload failed\",\n      error: inspect(metadata.error),\n      filename: metadata.filename\n    )\n  end\n\n  # Handle other events...\nend\n```","ref":"telemetry-and-monitoring.html#basic-setup"},{"type":"extras","title":"Metrics Collection - Telemetry and Monitoring","doc":"For detailed metrics and monitoring, integrate with your preferred metrics system.","ref":"telemetry-and-monitoring.html#metrics-collection"},{"type":"extras","title":"StatsD Integration - Telemetry and Monitoring","doc":"```elixir\ndefmodule MyApp.Metrics.StatsD do\n  def setup do\n    :telemetry.attach_many(\n      \"buckets-statsd\",\n      buckets_events(),\n      &handle_event/4,\n      nil\n    )\n  end\n\n  defp handle_event([:buckets, :cloud, event, :stop], measurements, metadata, _config) do\n    # Record timing\n    StatsD.histogram(\n      \"buckets.cloud.#{event}.duration\",\n      measurements.duration / 1_000_000,\n      tags: [\"adapter:#{metadata.adapter}\", \"cloud:#{cloud_name(metadata)}\"]\n    )\n\n    # Count operations\n    StatsD.increment(\n      \"buckets.cloud.#{event}.count\",\n      tags: [\"adapter:#{metadata.adapter}\"]\n    )\n  end\n\n  defp handle_event([:buckets, :cloud, event, :exception], _measurements, metadata, _config) do\n    StatsD.increment(\n      \"buckets.cloud.#{event}.error\",\n      tags: [\"adapter:#{metadata.adapter}\", \"error:#{error_type(metadata.error)}\"]\n    )\n  end\nend\n```","ref":"telemetry-and-monitoring.html#statsd-integration"},{"type":"extras","title":"Prometheus Integration - Telemetry and Monitoring","doc":"```elixir\ndefmodule MyApp.Metrics.Prometheus do\n  use Prometheus.Metric\n\n  def setup do\n    # Define metrics\n    Histogram.new(\n      name: :buckets_operation_duration_seconds,\n      help: \"Duration of bucket operations\",\n      labels: [:operation, :adapter, :cloud],\n      buckets: [0.01, 0.05, 0.1, 0.5, 1, 5, 10]\n    )\n\n    Counter.new(\n      name: :buckets_operations_total,\n      help: \"Total number of bucket operations\",\n      labels: [:operation, :adapter, :cloud, :status]\n    )\n\n    # Attach handlers\n    :telemetry.attach_many(\n      \"buckets-prometheus\",\n      buckets_events(),\n      &handle_event/4,\n      nil\n    )\n  end\n\n  defp handle_event([:buckets, :cloud, operation, :stop], measurements, metadata, _config) do\n    Histogram.observe(\n      [name: :buckets_operation_duration_seconds],\n      measurements.duration / 1_000_000_000,\n      labels: [operation, metadata.adapter, cloud_name(metadata)]\n    )\n\n    Counter.inc(\n      [name: :buckets_operations_total],\n      labels: [operation, metadata.adapter, cloud_name(metadata), \"success\"]\n    )\n  end\nend\n```","ref":"telemetry-and-monitoring.html#prometheus-integration"},{"type":"extras","title":"Custom Metrics - Telemetry and Monitoring","doc":"Add your own metrics around Buckets operations:\n\n```elixir\ndefmodule MyApp.Storage.Metrics do\n  def upload_with_metrics(file_path) do\n    start_time = System.monotonic_time()\n    file_size = File.stat!(file_path).size\n\n    metadata = %{\n      file_path: file_path,\n      file_size: file_size\n    }\n\n    :telemetry.span(\n      [:my_app, :storage, :upload],\n      metadata,\n      fn ->\n        object = Buckets.Object.from_file(file_path)\n        result = MyApp.Cloud.insert(object)\n        \n        {result, Map.put(metadata, :stored?, match?({:ok, _}, result))}\n      end\n    )\n  end\nend\n```","ref":"telemetry-and-monitoring.html#custom-metrics"},{"type":"extras","title":"Performance Monitoring - Telemetry and Monitoring","doc":"Track performance across different adapters and operations:\n\n```elixir\ndefmodule MyApp.PerformanceMonitor do\n  use GenServer\n\n  def start_link(_opts) do\n    GenServer.start_link(__MODULE__, %{}, name: __MODULE__)\n  end\n\n  def init(_state) do\n    :telemetry.attach_many(\n      \"performance-monitor\",\n      [\n        [:buckets, :adapter, :put, :stop],\n        [:buckets, :adapter, :get, :stop]\n      ],\n      &handle_event/4,\n      nil\n    )\n\n    {:ok, %{}}\n  end\n\n  defp handle_event([_, _, operation, :stop], measurements, metadata, _config) do\n    GenServer.cast(__MODULE__, {:record, operation, measurements, metadata})\n  end\n\n  def handle_cast({:record, operation, measurements, metadata}, state) do\n    key = {metadata.adapter, operation}\n    duration_ms = measurements.duration / 1_000_000\n\n    # Alert if slow\n    if duration_ms > 5000 do\n      Logger.warn(\"Slow #{operation} operation\",\n        adapter: metadata.adapter,\n        duration_ms: duration_ms,\n        path: metadata.path\n      )\n    end\n\n    # Update rolling average\n    new_state = Map.update(state, key, [duration_ms], fn durations ->\n      [duration_ms | Enum.take(durations, 99)]\n    end)\n\n    {:noreply, new_state}\n  end\nend\n```","ref":"telemetry-and-monitoring.html#performance-monitoring"},{"type":"extras","title":"Error Tracking - Telemetry and Monitoring","doc":"Monitor and alert on errors:\n\n```elixir\ndefmodule MyApp.ErrorTracker do\n  def setup do\n    :telemetry.attach(\n      \"error-tracker\",\n      [:buckets, :cloud, :insert, :exception],\n      &track_error/4,\n      nil\n    )\n  end\n\n  defp track_error(_event, _measurements, metadata, _config) do\n    Sentry.capture_exception(metadata.error,\n      stacktrace: metadata.stacktrace,\n      extra: %{\n        filename: metadata.filename,\n        cloud_module: metadata.cloud_module,\n        adapter: metadata.adapter\n      }\n    )\n  end\nend\n```","ref":"telemetry-and-monitoring.html#error-tracking"},{"type":"extras","title":"Dashboards - Telemetry and Monitoring","doc":"Create dashboards to visualize your storage metrics:","ref":"telemetry-and-monitoring.html#dashboards"},{"type":"extras","title":"Grafana Dashboard - Telemetry and Monitoring","doc":"Key panels to include:\n- Upload/download rate\n- Operation duration percentiles\n- Error rate by adapter\n- Storage usage trends\n- Concurrent operations","ref":"telemetry-and-monitoring.html#grafana-dashboard"},{"type":"extras","title":"Example Query (Prometheus) - Telemetry and Monitoring","doc":"```promql\n# Upload success rate\nrate(buckets_operations_total{operation=\"insert\",status=\"success\"}[5m])\n/\nrate(buckets_operations_total{operation=\"insert\"}[5m])\n\n# P95 upload duration\nhistogram_quantile(0.95,\n  rate(buckets_operation_duration_seconds_bucket{operation=\"insert\"}[5m])\n)\n```","ref":"telemetry-and-monitoring.html#example-query-prometheus"},{"type":"extras","title":"Best Practices - Telemetry and Monitoring","doc":"1. **Sample high-volume events** - Avoid overwhelming metrics systems\n2. **Use consistent labels** - Maintain standard label naming\n3. **Alert on anomalies** - Set up alerts for error spikes\n4. **Track business metrics** - Not just technical metrics\n5. **Monitor costs** - Track usage that affects billing\n6. **Regular review** - Analyze metrics for optimization opportunities","ref":"telemetry-and-monitoring.html#best-practices"},{"type":"extras","title":"Next Steps - Telemetry and Monitoring","doc":"- Implement [Security Monitoring](security-best-practices.html)\n- Set up [Performance Optimization](performance-optimization.html)\n- Configure [Production Monitoring](../deployment/production-checklist.html)","ref":"telemetry-and-monitoring.html#next-steps"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,49]}}